/************************************************************************/
/* mystic.h                                                             */
/*                                                                      */
/* MYSTIC - Monte Carlo code for the physically correct tracing of      */
/*          photons in cloudy atmospheres.                              */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*         DLR, bernhard.mayer@dlr.de                                   */
/*                                                                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 2000 Bernhard Mayer                                    */
/************************************************************************/

#ifndef __mystic_h
#define __mystic_h

#if defined (__cplusplus)
extern "C" {
#endif

#include <stdio.h>  /* required for FILENAME_MAX */

/* radiation source */
#define MCSRC_NONE                 0
#define MCSRC_SOLAR                1
#define MCSRC_ISOTROPIC_SURFACE    2
#define MCSRC_ISOTROPIC_TOP        3
#define MCSRC_THERMAL_SURFACE      4
#define MCSRC_THERMAL_ATMOSPHERE   5

/* albedo method */
#define MCALB_NONE     0
#define MCALB_LAM      1
#define MCALB_LAM2D    2
#define MCALB_RPV      3
#define MCALB_RPV2D    4


/* boundary conditions */
#define MCBCOND_NONE       0
#define MCBCOND_PERIODIC   1
#define MCBCOND_MIRROR     2


/* calculate 3D absorption field? */
#define MCABS_NONE        0    /* no                                */
#define MCABS_ABSORPTION  1    /* calculate absorption              */
#define MCABS_EMISSION    2    /* calculate emission                */
#define MCABS_HEATING     3    /* calculate (absorption - emission) */
#define MCABS_ACTINIC     4    /* calculate actinic flux            */

/* azimuth conventions */
#define MCAZIMUTH_NEW 0
#define MCAZIMUTH_OLD 1


typedef enum {NONE, UPPER, LOWER} BOUNDARY;

typedef struct {
  double a;
  double b;
  double c;
  double d;
  double upper;
} surface;


typedef struct {
  int n;
  double *air;
  double *aer;
  double *ozo;
  double *cld;
  double *ice;
  double *tot;
} profile;


/* profile3D is required to store the profiles of  */
/* absorption and scattering coefficients; float   */
/* should be accurate enough                       */

typedef struct {
  int Nz;
  int Nx;
  int Ny;
  int *threed;     /* flag if a layer is 3D */
  int nthreed;     /* number of 3D layers   */
  float ***cld;
  float ***tot;
} profile3D;


/* ddprofile3D is required to store the profile of */
/* the absorbed energy; as all output quantities,  */
/* this parameter is also stored in double rather  */
/* than float                                      */

typedef struct {
  int Nz;
  int Nx;
  int Ny;
  int *threed;
  int nthreed;
  /* double ***aer; */
  /* double ***ozo; */
  /* double ***ice; */
  double ***cld;
  double ***tot;
} ddprofile3D;


/* table of integrated phase function, equidistant in F; */
/* used to calculate mu for a given F.                   */
typedef struct {
  double *F;
  double *D;
  double *mu;
  int nF;
} ipft;

/* table of phase function, equidistant in mu; */
/* used to calculate p for a given mu.         */
typedef struct {
  double *p;
  double *mu;
  double dmu1;
  double dmu2;

  int truncate;    /* truncation flag */
  double mucut;    /* truncation mu */
  double a;        /* linear replacement of forward peak, intersect */
  double b;        /* linear replacement of forward peak, slope     */

  int np1;
  int np2;
} pft;

typedef struct {
  ipft **iphase;   /* integrated phase function table for each effective radius reff[i] */
  pft  **phase;    /* phase function iphase for each effective radius reff[i]           */
  double *ext;     /* extinction coefficient for each effective radius                  */
  double *dext;    /* extinction coefficient increment                                  */
  double *omega;   /* single scattering albedo for each effective radius                */
  double *domega;  /* single scattering albedo increment                                */
  double r0;       /* reff[i] = r0 + i * dr; i = 0 ... n-1                              */
  double dr;
  int    n;
} phase_function_table_array;


typedef struct {
  ipft *iphase;   /* integrated phase function table */
  pft  *phase;    /* phase function iphase           */
} phase_function_table;


typedef struct {
  double air;
  double aer;
  double ozo;
  double cld;
  double ice;
  double tot;
} optical_depth;


typedef struct {
  int    **ndir;    /* number of photons */
  int    **ndn;
  int    **nup;

  double **edir;    /* irradiance        */
  double **edn; 
  double **eup;  

  double **fdir;    /* actinic flux      */
  double **fdn; 
  double **fup;  

  double ***raddir; /* radiance          */
  double ***raddif;
  double ***radesc;
} radiation_field;


typedef struct {
  radiation_field *surf;
  radiation_field **alt;
  ddprofile3D *absorption3D;
} result_struct;


typedef struct {
  double dx[3];
  int hit[3];
  double cotheta; 
} direction;


/* Structure to store a direction and opening angle */
/* for radiance calculations; convention:           */ 
/*    phi=0,   South                                */
/*    phi=90,  West                                 */
/*    phi=180, North                                */
/*    phi=270, East                                 */
/*                                                  */
/*    theta > 0 upward                              */
/*    theta < 0 downward                            */

typedef struct {
  double theta;         /* polar angle                     */
  double phi;           /* azimuth, internal               */
  double externalphi;   /* azimuth, as defined by the user */
  double alpha; 
  double cosalpha; 
  double omega; 
  direction dir;
} radang;


typedef struct {
  double x[3];          /* location                                 */
  direction dir;        /* direction                                */
  int direct;           /* direct or diffuse                        */
  int scattercounter;   /* number of scattering + reflection events */ 

  double pathlength;    /* photon pathlength                      */
  optical_depth tauabs; /* absortion optical depth along the path */

  BOUNDARY boundary;    /* flag: hit lower 1D boundary yes/no  */
  int boundary2D;       /* flag: hit lower 3D boundary yes/no  */
  int crossingboundary; /* flag: left current box yes/no       */
  int reallyhit[3];     /* flag: crossed box boundary          */

  int ic;               /* box index, x */
  int jc;               /* box index, y */
  int kc;               /* box index, z */

  double *pdir;         /* probability for direction radang[]  */

  int ipa;              /* independent pixel approximation     */
} photon_struct;


typedef struct {
  double k;
  double rho0;
  double theta;

  double dmu;        
  double *alb;  /* albedo[i] for incidence angle mu = i * dmu, i = 0 ... n-1 */
  double *dalb;  
  int n;
} rpv_struct;


typedef struct {
  int method;
  char filename[FILENAME_MAX];

  double **albedo2D;
  double albedo;
  double delX;
  double delY;
  int Nx;
  int Ny;
  double Wsurf;

  char rpv_filename [FILENAME_MAX];
  rpv_struct *rpv;   /* rpv structure                     */
  int *rpv_count;    /* # of occurrence of each RPV index */

  char **rpv_index;  /* xy-field of RPV indices           */
} albedo_struct;


typedef struct {
  int elev2D;
  char filename[FILENAME_MAX];

  surface **surf;
  double surfmax;
  double delX;
  double delY;
  int Nx;
  int Ny;
} elevation_struct;

typedef struct {
  double delX;
  double delY;
  double *X;
  double *Y;
  double *Z;
  int Nx;
  int Ny;
  int Nz;
  double xmax;
  double ymax;
  int *threed;
  int nthreed;

  char wcloudfilename [FILENAME_MAX];

  profile *dtsca;
  profile *dtabs;
  profile *ksca;
  profile *kabs;
  profile *g1;
  profile *g2;
  profile *ff;
  profile *lwc;
  profile *reff;
  profile *Bplanck;

  profile3D *ksca3D;
  profile3D *kabs3D;

  profile3D *g1cld3D;
  profile3D *g2cld3D;
  profile3D *ffcld3D;

  profile3D *lwc3D;
  
  double gmin;
  double gmax;
  int gspec;

  optical_depth tabstot;
  optical_depth tscatot;

  double Watm;
  double maxemis;
} atmosphere_struct;


typedef struct {
  int *sample;
  radang *rad;
  int nrad;
  double delX;
  double delY;
  int Nx;
  int Ny;
  int surface;
  int escape;
  int passback3D;
} sample_struct;



/* prototypes */
int mystic (int *nlyr,
	    float *dtair, float *dto3, 
	    float *dtaer, float *omaer, float *g1aer, float *g2aer, float *faer,
	    float *dtcld, float *omcld, float *g1cld, float *g2cld, float *fcld,
	    float *dtice, float *omice, float *g1ice, float *g2ice, float *fice,
	    float *lwccld, float *recld,
	    float *temper, 
	    float *zprof,
	    float *btemp,
	    float *umu, int *numu,
	    float *phi, int *nphi,
	    int aziconv,
	    char  *momcld_filename,
	    char  *pcld_filename,
	    float **momaer, int *nmomaer,
	    float *alb, float *sza, float *phi0, int *mcphotons,
	    int *source, float *wvnmlo, float *wvnmhi,
	    int *nxout, int *nyout, float *zout, int *nzout,
	    float *rpv_rho0, float *rpv_k, float *rpv_theta, 
	    int *ipa, int *escape, int *absorption, int *loaddata,
	    char *basename, 
	    char *cloudfilename,
	    char *albedofilename,
	    char *rpvfilename, char **rpv_labels, int *rpv_nlabels,
	    char *elevationfilename,
	    float *truncate,
	    int *bcond,
	    int *quiet,
	    float *rfldir, float *rfldn,  float *flup,
	    float *uavgso, float *uavgdn, float *uavgup,
	    float *rfldir3d, float *rfldn3d, float *flup3d, float *radiance3d,
	    float *abs3d, int *nxcld, int *nycld, int *nzcld);
  
double sc_Isotropic_phi();
double sc_Rayleigh_theta();
double sc_Isotropic_theta();
double sc_Lambertian_theta();
double sc_HG_theta(double g);
double sc_HG2_theta(double g1, double g2, double ff);
double random_tau();
  

double rpv_brdf (double rho0, double k, double theta, 
		 double mu1, double mu2, double phi);

double rpv_albedo (double rho0, double k, double theta, double mu1);

int random_rpv_direction (double rho0, double k, double theta, double mu1,
			  double *mu2, double *phi);


#if defined (__cplusplus)
}
#endif

#endif
