/*--------------------------------------------------------------------
 * $Id: aerosol.c,v 1.26 2004/07/24 10:29:42 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/numeric.h"
#include "../src_c/miecalc.h"
#include "../src_c/fortran_and_c.h"
#include "../src_f/f77-uscore.h"

#define EPSILON 1E-6

/************************************/
/* prototypes of internal functions */
/************************************/

static int read_aerosol_profile (char *filename, int nlyr, int nlambda_t, 
				 float *zd_common, float **dtau, int scale);

static int read_aerosol_moments (char *filename,
				 float **moment, int *nmom);

static int scale_aer_tau (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			  float tau, float *zd, float altitude);

static int scale_aer_angstrom (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			       float alpha, float beta, float *zd, float altitude);

static int aerosol_set_visibility (aer_inp_struct inp, char *path, float *visib, int verbose);



/**************************************************************/
/* Read file holding  filenames for each layer.               */
/**************************************************************/

int read_optprop_files (char *filename,
			float *lambda_t, int nlambda_t, 
			int nlambda_rte_lower, int nlambda_rte_upper,
			float *zd, int nlev, 
			float **dtau, float **ssa, float **gg,
			float ***mom, int **nmom, int verbose) 
{
  int i=0, iv=0, k=0, lc=0;
  int status=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;
  int arows=0, amin_columns=0, amax_columns=0;
  int tmp_nlambda=0, tmp_nmom=0;
  int linear=1;
  float *tmp_lambda=NULL;
  float *tmp_dtau=NULL, *tmp_ssa=NULL, *tmp_gg=NULL;
  float **tmp_mom=NULL;
  float **data=NULL;
  float *temp=NULL, **temp_mom=NULL;
  char ***string=NULL;
  char *dummy=NULL;
  int nonzero=0;

  if ((status = ASCII_checkfile (filename, &rows, &min_columns, &max_columns, &max_length)) != 0) 
    return status;

  if (min_columns<2) {
    fprintf (stderr, "Error, need at least 2 columns in optical properties file %s\n", filename);
    return -1;
  }

  if ((status = ASCII_calloc_string(&string, rows, max_columns, FILENAME_MAX+1)) != 0) 
    return status;

  if ((status = ASCII_readfile(filename, string)) != 0) 
    return status;

  if (nlev != rows) {
    fprintf (stderr, "Fatal programming error in read_optprop_files(); nlev=%d, rows=%d\n",
	     nlev, rows);
    return -1;
  }
  

  /* copy altitude grid */
  for (i=0; i<=nlev-1; i++)
    zd[i] = (float) strtod (string[i][0], &dummy);
  
  
  temp = calloc (nlambda_t, sizeof(float));
					   
  /* check if the optical properties file of the              */
  /* uppermost layer has zero optical thickness;              */
  /* we need that one to define the upper layer boundary for  */
  /* the uppermost aerosol/cloud layer; as an alternative     */
  /* NULL as filename simply says that the uppermost layer    */
  /* is empty                                                 */                                               

  if (strcasecmp(string[0][1], "NULL")) {  /* if the uppermost string is not NULL */
    
    if ((status = ASCII_file2float (string[0][1], &arows, &amax_columns,
				    &amin_columns, &data)) != 0) {
      fprintf (stderr, "Error %d reading file %s for layer %d.\n", 
	       status, string[0][1], lc);
      return status;
    }
    
    if (amin_columns<4) {
      fprintf (stderr, "Error, need at least 4 columns in optical properties file %s\n", 
	       string[0][1]);
      return -1;
    }
    
    for (iv=0; iv<arows; iv++)
      if (data[iv][1] > 0)  /* optical thickness is first column */
	nonzero++;
    
    if (nonzero) {
      fprintf (stderr, "Error, need a layer with zero optical thickness as uppermost\n");
      fprintf (stderr, "layer in %s. Please add one and retry!\n", string[0][1]);
      fprintf (stderr, "See examples/AERO_FILES as an example!\n");
      return -1;
    }
    
    ASCII_free_float (data, arows);
  }
  else {
    fprintf (stderr, "  ... found NULL as uppermost layer, fine :-)\n");
  }

  /* For each layer read the optical properties file.                  */
  /* Note that the wavelength resolution may be different from the one */
  /* in the extraterrestrial solar flux file.                          */
  for (lc=0; lc<nlev-1; lc++) {

    if (verbose)
      fprintf (stderr, " ... reading optical properties %s for layer [%.4f, %.4f]\n",
	       string[lc+1][1], zd[lc+1], zd[lc]);

    if ((status = ASCII_file2float (string[lc+1][1], &arows, &amax_columns,
				    &amin_columns, &data)) != 0) {
      fprintf (stderr, "Error %d reading file %s for layer %d.\n", 
	       status, string[lc+1][1], lc);
      return status;
    }
    
    if (amin_columns<4) {
      fprintf (stderr, "Error, need at least 4 columns in optical properties file %s\n", 
	       string[lc+1][1]);
      return -1;
    }
    
    tmp_nmom = amax_columns-3;
    tmp_nlambda = arows;
      
    tmp_lambda = calloc (tmp_nlambda, sizeof(float));
    tmp_dtau   = calloc (tmp_nlambda, sizeof(float));
    tmp_gg     = calloc (tmp_nlambda, sizeof(float));
    tmp_ssa    = calloc (tmp_nlambda, sizeof(float));
      
    if ((ASCII_calloc_float (&tmp_mom,  tmp_nmom, tmp_nlambda)) != 0)
      return status;

    for (iv=0; iv<tmp_nlambda; iv++)
      tmp_lambda[iv] = data[iv][0];

    for (iv=0; iv<tmp_nlambda; iv++) {
      tmp_dtau [iv] = data[iv][1];	 /* Still in units of km-1      */
      tmp_ssa  [iv] = data[iv][2];	
      for (k=0; k<tmp_nmom; k++)
	if (data[iv][k+3]==data[iv][k+3]) /* Trick to replace NaN by 0  */
	  tmp_mom[k][iv] = data[iv][k+3];
	else
	  tmp_mom[k][iv] = 0;

      if (tmp_nmom>=2)
	tmp_gg [iv] = tmp_mom[1][iv];
      else 
	tmp_gg [iv] = 0;
    }

    ASCII_free_float (data, arows);

    for (iv=0; iv<tmp_nlambda; iv++)
      tmp_dtau[iv] *= (zd[lc]-zd[lc+1]);
    
    /* Interpolate from optical property wavelengths to */
    /* the internal transmittance wavelength grid.      */
    
    /* interpolate optical thickness */
    status = arb_wvn2 (tmp_nlambda, tmp_lambda, tmp_dtau,  
		       nlambda_rte_lower, nlambda_rte_upper,
		       lambda_t, temp, linear);

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating dtau in read_optprop_files()\n", status);
      return status;
    }

    for (iv=0; iv<nlambda_t; iv++)
      dtau[iv][lc] = temp[iv];

    /* interpolate asymmetry parameter */
    status = arb_wvn2 (tmp_nlambda, tmp_lambda, tmp_gg,  
		       nlambda_rte_lower, nlambda_rte_upper,
		       lambda_t, temp, linear);

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating gg in read_optprop_files()\n", status);
      return status;
    }

    for (iv=0; iv<nlambda_t; iv++) 
      gg[iv][lc] = temp[iv];
    

    /* interpolate single scattering albedo */
    status = arb_wvn2 (tmp_nlambda, tmp_lambda, tmp_ssa,
		       nlambda_rte_lower, nlambda_rte_upper,
		       lambda_t, temp, linear);
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating ssa in read_optprop_files()\n", status);
      return status;
    }
    
    for (iv=0; iv<nlambda_t; iv++) 
      ssa[iv][lc] = temp[iv];

    

    ASCII_calloc_float (&temp_mom, tmp_nmom, nlambda_t);

    for (k=0; k<tmp_nmom; k++) {
      status = arb_wvn2 (tmp_nlambda, tmp_lambda, tmp_mom[k],
			 nlambda_rte_lower, nlambda_rte_upper,
			 lambda_t, temp_mom[k], linear);
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating mom in read_optprop_files()\n", status);
	return status;
      }
    }

    /* get index of maximum nonzero moment for each wavelength */
    for (iv=0; iv<nlambda_t; iv++) {
      for (k=tmp_nmom-1; k>=0; k--)  
	if (temp_mom[k][iv]!=0)
	  break;
      
      nmom[iv][lc] = k+1;
      
      mom[iv][lc] = calloc (nmom[iv][lc], sizeof(float));
      for (k=0; k<nmom[iv][lc]; k++)
	mom[iv][lc][k] = temp_mom[k][iv];
    }
      

    /* free memory */
    ASCII_free_float (temp_mom, tmp_nmom);
    ASCII_free_float (tmp_mom,  tmp_nmom);

    free (tmp_dtau);
    free (tmp_gg);
    free (tmp_ssa);
    free (tmp_lambda);
  }

  ASCII_free_string (string, rows, max_columns);
  free (temp);
  
  return 0;
}



/**************************************************************/
/* Setup atmospheric aerosol.                                 */
/**************************************************************/

int setup_aerosol (input_struct input, output_struct *output)
{
  int status=0, ip=0, iv=0, lc=0, k=0, kmax=0, nlyr=0, ialt=0;
  int maxawvn = MAXAWVN;

  int linear=0;
  int refrac=0, sizedist=0;

  int rows=0, min_columns=0, max_columns=0, max_length=0;

  float *tmp_aerext=NULL, *tmp_aerabs=NULL, *tmp_aersym=NULL, *tmp_aerwvn=NULL;  
  float *tmp_aer_dtau=NULL, *tmp_aer_gg=NULL, *tmp_aer_ssa=NULL;
  float *tmp_rh=NULL;

  float *lref=NULL, *re=NULL, *im=NULL;
  int nref=0;

  double temperature = 300.0;  /* The temperature is only used for wavelengths > 167 um */
  double beta=0, beta0=0, ext=0, omega=0, g=0;
  
  double *tautot = NULL, *lambda = NULL;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double tau550=0, scale_factor=0;

  mie_complex crefin = {0.0, 0.0}, ref = {0.0, 0.0};

  mie_inp_struct *mie_inp=NULL;
  mie_out_struct *mie_out=NULL;

  float *moment=NULL;

  float *tmp_zd=NULL;
  int tmp_nlev=0;

  aer_out_struct *tmpaer = calloc (1, sizeof(aer_out_struct));
  
  int nmom=1000;   /* ??? should tie this to nstr to save computational time ??? */

  float *zalt=NULL;


  void F77_FCN (aerprof) (int *nlyr, float *zd, float *aer_pro, 
			  int *seasn, int *vulcan, float *visib );
  void F77_FCN (aeropt) ( int *vulcan, float *aerext, float *aerabs, 
			  float *aersym, float *aerwvn, float *rh, 
			  int *nlyr, float *zd, int *haze, int *maxawvn ); 
  void F77_FCN (aerwvn) ( float *aerwvn, float *aerabs, float *aersym, float *aerext,
			  float *aer_pro, float *aer_dtau, float *aer_gg, 
			  float *aer_ssa, int *nlyr, float *lambda_t, float *zd, 
			  int *nlambda_t, int *maxawvn );


  if (input.aer.vulcan > 0) {
    /* first, determine the complete set of altitudes   */
    /* for the aerosol; to do that, combine the z-grids */
    /* of all profiles                                  */
    
    /* we want all atmospheric levels in the aerosol profile  */
    /* because the default aerosol (Shettle) is defined       */
    /* on the atmospheric grid; user-defined levels are       */
    /* required as well, hence we use the common grid as      */
    /* input which contains all atmospheric levels as well as */
    /* the user-defined ones                                  */
    
    output->aer.nlev = 0; 
    set_common_z_grid (output->atm.zd_common, output->atm.nlev_common, 
		       &tmp_zd, &tmp_nlev);
    
    status=0;
    if (strlen(input.aer.filename[FN_AER_FILES]) > 0)
      status += add_file2grid (&tmp_zd, &tmp_nlev, input.aer.filename[FN_AER_FILES]);
    
    if (strlen(input.aer.filename[FN_AER_TAU]) > 0)
      status += add_file2grid (&tmp_zd, &tmp_nlev, input.aer.filename[FN_AER_TAU]);
    
    if (strlen(input.aer.filename[FN_AER_GG]) > 0)
      status += add_file2grid (&tmp_zd, &tmp_nlev, input.aer.filename[FN_AER_GG]);
    
    if (strlen(input.aer.filename[FN_AER_SSA]) > 0) 
      status += add_file2grid (&tmp_zd, &tmp_nlev, input.aer.filename[FN_AER_SSA]);
    
    if (status!=0) {
      fprintf (stderr, "Error creating combined aerosol grid\n");
      return status;
    }
    
    /* allocate memory for profiles */
    status = calloc_aer_out (&(output->aer), output->wl.nlambda_t, tmp_nlev);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by calloc_aer_out()\n", status);
      return status;
    }
    
    /* copy levels */
    for (lc=0; lc<output->aer.nlev; lc++)
      output->aer.zd[lc] = tmp_zd[lc];
    
    free(tmp_zd);
    

    /* the Shettle aerosol profile is defined starting       */
    /* at the surface not at z=0; need to subtract altitude  */
    zalt = calloc (output->aer.nlev, sizeof(float));
    for (lc=0; lc<output->aer.nlev; lc++)
      zalt[lc] = output->aer.zd[lc] - output->atm.altitude;

    /* copy visibility to destination and change to a reasonable value if needed */
    status = aerosol_set_visibility (input.aer, input.filename[FN_PATH], &(output->aer.visibility), input.verbose);

    /**** Get appropriate aerosol profile and interpolate profile to zd grid ****/

    nlyr = output->aer.nlev-1;

    F77_FCN (aerprof) (&nlyr, zalt, output->aer.aer_pro, 
		       &input.aer.seasn, &input.aer.vulcan, &(output->aer.visibility));
    
    /**** Get appropriate aerosol optical properties ****/
    
    tmp_aerext = (float *) calloc (maxawvn * output->aer.nlev, sizeof(float));
    tmp_aerabs = (float *) calloc (maxawvn * output->aer.nlev, sizeof(float));
    tmp_aersym = (float *) calloc (maxawvn * output->aer.nlev, sizeof(float));
    tmp_aerwvn = (float *) calloc (maxawvn * output->aer.nlev, sizeof(float));
    
    /* interpolate relative humidity linearely to aerosol grid */
    tmp_rh = calloc (output->aer.nlev, sizeof (float));
    status = arb_wvn (output->atm.nlev, output->atm.zd, output->atm.microphys.rh,
		      output->aer.nlev, output->aer.zd,  tmp_rh, 1, 1);

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating relative humidity to aerosol profile\n", status);
      return status;
    }

    F77_FCN (aeropt) (&input.aer.vulcan, tmp_aerext, tmp_aerabs, 
		      tmp_aersym, tmp_aerwvn, tmp_rh, 
		      &nlyr, zalt, &input.aer.haze, &maxawvn); 
    
    free (tmp_rh);

    tmp_aer_dtau = (float *) calloc (output->wl.nlambda_t * nlyr, sizeof(float));
    tmp_aer_gg   = (float *) calloc (output->wl.nlambda_t * nlyr, sizeof(float));
    tmp_aer_ssa  = (float *) calloc (output->wl.nlambda_t * nlyr, sizeof(float));
    
    F77_FCN (aerwvn) (tmp_aerwvn, tmp_aerabs, tmp_aersym, tmp_aerext, output->aer.aer_pro, 
		      tmp_aer_dtau, tmp_aer_gg, tmp_aer_ssa, &nlyr,
		      output->wl.lambda_t, zalt, &output->wl.nlambda_t, &maxawvn);
    
    fortran2c_2D_float_ary_noalloc (output->wl.nlambda_t, nlyr, tmp_aer_gg,   output->aer.optprop.g1);
    fortran2c_2D_float_ary_noalloc (output->wl.nlambda_t, nlyr, tmp_aer_ssa,  output->aer.optprop.ssa);
    fortran2c_2D_float_ary_noalloc (output->wl.nlambda_t, nlyr, tmp_aer_dtau, output->aer.optprop.dtau);
    
    free(tmp_aer_gg);
    free(tmp_aer_ssa);
    free(tmp_aer_dtau);
    
    free(tmp_aerext);
    free(tmp_aerabs);
    free(tmp_aersym);
    free(tmp_aerwvn);



    /**** Read file holding filenames of layer aerosol ****/
    /**** optical properties if specified              ****/

    if (strlen(input.aer.filename[FN_AER_FILES]) > 0) {
      
      /* determine number of levels; required for memory allocation */
      if ((status = ASCII_checkfile (input.aer.filename[FN_AER_FILES], 
				     &rows, &min_columns, &max_columns, &max_length)) != 0) {
	fprintf (stderr, "Error %d reading %s\n", status, input.filename[FN_AER_FILES]);
	return status;
      }
      
      /* allocate memory for temporary array */
      status = calloc_aer_out (tmpaer, output->wl.nlambda_t, rows);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for aer_out struct\n", status);
	return status;
      }

      /* then read optical properties */
      status = read_optprop_files (input.aer.filename[FN_AER_FILES],
				   output->wl.lambda_t, output->wl.nlambda_t, 
				   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				   tmpaer->zd, tmpaer->nlev, 
				   tmpaer->optprop.dtau, tmpaer->optprop.ssa, tmpaer->optprop.g1,
				   tmpaer->optprop.moment, tmpaer->optprop.nmom, input.verbose);

      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol optical property files: %s\n", 
		 status, input.aer.filename[FN_AER_FILES]);
	return status;
      }
      
      /* redistribute optical properties to common aerosol grid */
      status = redistribute_optprop (&(tmpaer->optprop), output->wl.nlambda_t, 
				     output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				     tmpaer->zd, tmpaer->nlev-1, 
				     output->aer.zd, output->aer.nlev-1);
  
      /* re-write z profile */
      free (tmpaer->zd);
      tmpaer->nlev = output->aer.nlev;
      tmpaer->zd = calloc (tmpaer->nlev, sizeof(float));
      for (lc=0; lc<output->aer.nlev; lc++)
	tmpaer->zd[lc] = output->aer.zd[lc];

      if (status!=0) {
	fprintf (stderr, "Error regridding aerosol parameters\n");
	return status;
      }
      
      /* copy common aerosol grid to final destination */
      calloc_aer_out (&(output->aer), output->wl.nlambda_t, output->aer.nlev);
      status = cp_aer_out (tmpaer, &(output->aer), output->wl.nlambda_t, 1, input.quiet);
      if (status!=0) {
	fprintf (stderr, "Error %d copying temporary aerosol properties to\n", status);
	fprintf (stderr, "output->aer\n");
	return status;
      }


      /* increase number of moments if necessary */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->aer.nlev-1; lc++)
	  if (output->aer.optprop.nmom[iv][lc]-1 > output->atm.nmom)
	    output->atm.nmom = output->aer.optprop.nmom[iv][lc]-1;
    }

    
    /****   Read aerosol optical depth file if specified ****/
    if (strlen(input.aer.filename[FN_AER_TAU]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_TAU], output->aer.nlev-1,
				     output->wl.nlambda_t, output->aer.zd,
				     output->aer.optprop.dtau, 1);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol tau file: %s\n", 
		 status, input.aer.filename[FN_AER_TAU]);
	return status;
      }
    }


    /****   Read aerosol asymmetry parameter file if specified ****/
    if (strlen(input.aer.filename[FN_AER_GG]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_GG], output->aer.nlev-1,
				     output->wl.nlambda_t, output->aer.zd,
				     output->aer.optprop.g1, 0);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol asymmetry parameter file: %s\n", 
		 status, input.aer.filename[FN_AER_GG]);
	return status;
      }
    }


    /****   Read aerosol single scattering albedo file if specified ****/
    if (strlen(input.aer.filename[FN_AER_SSA]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_SSA], output->aer.nlev-1,
				     output->wl.nlambda_t, output->aer.zd,
				     output->aer.optprop.ssa, 0);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol single scattering albedo file: %s\n", 
		 status, input.aer.filename[FN_AER_SSA]);
	return status;
      }
    }
    
    
    /**** Read aerosol moments file if specified ****/
    /**** At present: independent of wavelength  ****/
    /**** but this can be easily changed!        ****/
    
    if (strlen(input.aer.filename[FN_AER_MOM]) > 0) {
      
      status = read_aerosol_moments (input.aer.filename[FN_AER_MOM], 
				     &moment, &rows);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol moments file: %s\n", 
		 status, input.aer.filename[FN_AER_MOM]);
	return status;
      }
      
      /* copy to other wavelengths and layers */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->aer.nlev-1; lc++) {
	  output->aer.optprop.nmom[iv][lc] = rows;
	  output->aer.optprop.moment[iv][lc] = calloc (output->aer.optprop.nmom[iv][lc], sizeof (float));

	  for (k=0; k<output->aer.optprop.nmom[iv][lc]; k++)
	    output->aer.optprop.moment[iv][lc][k] = moment[k];
	}

      /* increase number of moments if more found */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->aer.nlev-1; lc++)
	  if (output->aer.optprop.nmom[iv][lc]-1 > output->atm.nmom)
	    output->atm.nmom = output->aer.optprop.nmom[iv][lc]-1;
      

      free (moment);
    }

    /****   Read aerosol refractive index   ****/
    if (strlen(input.aer.filename[FN_AER_REF]) > 0) {

      /* read wavelength-dependent aerosol refractive index */
      status = read_3c_file_float (input.aer.filename[FN_AER_REF], 
				   &lref, &re, &im, &nref);
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading aerosol refractive index file\n", status);
	return status;
      }

      /* linear interpolation of the refractive index to the model wavelengths */
      linear = 1;

      output->aer.re_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      output->aer.im_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

      status = arb_wvn (nref, lref, re, 
			output->wl.nlambda_t, output->wl.lambda_t, output->aer.re_t, 
			linear, 0);

      if (status != 0) {
	fprintf (stderr, "Error, status %d returned by arb_wvn()\n", status);
	return status;
      }

      status = arb_wvn (nref, lref, im, 
			output->wl.nlambda_t, output->wl.lambda_t, output->aer.im_t, linear, 0);

      if (status != 0) {
	fprintf (stderr, "Error, status %d returned by arb_wvn()\n", status);
	return status;
      }
      
      free(lref); free(re); free(im);

      refrac=1;
    }
    else {
      if (input.aer.re>0) {   /* constant refractive index */
	output->aer.re_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
	output->aer.im_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

	for (iv=0; iv<output->wl.nlambda_t; iv++) { 
	  output->aer.re_t[iv] = input.aer.re;
	  output->aer.im_t[iv] = input.aer.im;
	}
	refrac=1;
      }
    }


    /****   Read aerosol size distribution  ****/
    if (strlen(input.aer.filename[FN_AER_SIZ]) > 0) {

      /* read size distribution */
      status = read_2c_file (input.aer.filename[FN_AER_SIZ], 
			     &(output->aer.xsiz), &(output->aer.ysiz), &(output->aer.nsiz));
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading aerosol size distribution file\n", status);
	return status;
      }
      

      if (output->aer.xsiz[0]==0) {
	output->aer.xsiz[0] = 0.1 * output->aer.xsiz[1];
	if (!input.quiet)
	  fprintf (stderr, " ... aerosol size distribution: setting first data point to r=%g\n",
		   output->aer.xsiz[0]);
      }

      sizedist=1;
    }

    if (refrac&&sizedist) {

      if (!input.quiet)
	fprintf (stderr, " ... now doing aerosol Mie calculation\n");

      mie_inp = calloc (1, sizeof(mie_inp_struct));
      mie_out = calloc (1, sizeof(mie_out_struct));

      mie_out->pmom = calloc (4, sizeof(float *));
      for (ip=0; ip<4; ip++)
	mie_out->pmom[ip] = calloc (nmom+1, sizeof(float));
      

      /**********************************************************/
      /* The wavelength dependence of the optical thickness     */
      /* follows Mie theory; the absolute value of the optical  */
      /* thickness is defined by its value at this place of the */
      /* code; that is, the value defined by 'visibility' or    */
      /* (much better) by aerosol_set_tau. Later, it may still  */
      /* be scaled with the Angstrom parameters (if anybody     */
      /* wants that.						*/
      /* ??? I guess there are much better ways to do that ???  */
      /**********************************************************/

      for (iv=0; iv<output->wl.nlambda_t; iv++) {


	mie_inp->anyang      = 0;
	mie_inp->ipolzn      = 0;
	mie_inp->nmom        = nmom;
	mie_inp->momdim      = nmom+1;
	mie_inp->numang      = 0;
	mie_inp->mimcut      = 1.0e-08;
	mie_inp->perfct      = 0;
	mie_inp->prnt[0]     = 0;
	mie_inp->prnt[1]     = 0;
	mie_inp->xmu         = (float *)   calloc (mie_inp->numang+1, sizeof(float));
	mie_inp->s1          = (mie_complex *) calloc (mie_inp->numang+1, sizeof(mie_complex));
	mie_inp->s2          = (mie_complex *) calloc (mie_inp->numang+1, sizeof(mie_complex));


	crefin.re = output->aer.re_t[iv];
	crefin.im = output->aer.im_t[iv];

	status = mie_calc_sizedist (*mie_inp, mie_out, 
				    MIEV0, USER, crefin, 
				    (output->wl.lambda_t[iv])/1000.0, temperature,
				    output->aer.xsiz, output->aer.ysiz, output->aer.nsiz, 
				    &beta, &omega, &g, &ref);


	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mie_calc_sizedist()\n", status);
	  return status;
	}

	/* determine maximum non-zero moment */
	kmax=0;
	for (k=1; k<=nmom; k++)
	  if (mie_out->pmom[0][k] != 0)
	    kmax=k;
	
	for (lc=0; lc<output->aer.nlev-1; lc++)
	  output->aer.optprop.nmom[iv][lc] = kmax+1;

	/* copy moments to final destination */
	for (lc=0; lc<output->aer.nlev-1; lc++) {
	  output->aer.optprop.moment[iv][lc] = calloc (output->aer.optprop.nmom[iv][lc], sizeof(float));
	  for (k=0; k<kmax; k++)
	    output->aer.optprop.moment[iv][lc][k] = mie_out->pmom[0][k]/mie_out->pmom[0][0];
	}
	

	/* extinction at first wavelength; required for scaling the optical thickness */
	if (iv==0)
	  beta0 = beta;


	ext = 0.0;
	  
	/* overwrite optical thickness, single-scattering albedo, and asymmetry parameter */
	for (lc=0; lc<output->aer.nlev-1; lc++) {
	  output->aer.optprop.dtau [iv][lc] *= beta/beta0;
	  output->aer.optprop.ssa  [iv][lc]  = omega;
	  output->aer.optprop.g1   [iv][lc]  = g;
	  
	  ext += output->aer.optprop.dtau[iv][lc];
	}
	

	if (!input.quiet) {
	  if (iv==0)
	    fprintf (stderr, "mie  lambda  nmom       tau       ssa       asy\n");
	  
	  fprintf (stderr, "mie %7.2f  %4d  %g  %g  %g\n",
		   output->wl.lambda_t[iv], kmax, ext, omega, g);
	}
      }

      /* free Mie memory */
      for (ip=0; ip<4; ip++)
	free (mie_out->pmom[ip]);
      free(mie_out->pmom);

      free (mie_inp);
      free (mie_out);
    }


    /**** Set aerosol asymmetry factor ****/
    if (input.aer.set_gg >= -1.0)
      for (lc=0; lc<output->aer.nlev-1; lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++)
	  output->aer.optprop.g1[iv][lc] = input.aer.set_gg;
    
    /**** Set aerosol single scattering albedo ****/
    if (input.aer.set_ssa >= 0.0)
      for (lc=0; lc<output->aer.nlev-1; lc++)
	for (iv=0; iv<output->wl.nlambda_t; iv++)
	  output->aer.optprop.ssa[iv][lc] = input.aer.set_ssa;
    
    /**** Scale aerosol single scattering albedo ****/
    if (input.aer.scale_ssa >= 0.0)
      for (lc=0; lc<output->aer.nlev-1; lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++) {
	  output->aer.optprop.ssa[iv][lc] = output->aer.optprop.ssa[iv][lc] * input.aer.scale_ssa;
	  if (output->aer.optprop.ssa[iv][lc]>1.0) 
	    output->aer.optprop.ssa[iv][lc] = 1.0;
	}

    /**** Set aerosol optical thickness ****/
    if (input.aer.set_tau >= 0.0) {
      status = scale_aer_tau (output->aer.optprop.dtau, output->wl.lambda_t, output->wl.nlambda_t, 
			      output->aer.nlev-1, input.aer.set_tau, 
			      output->aer.zd, output->atm.altitude);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by scale_aer_tau()\n", status);
	return status;

      }
    }

    /**** Scale aerosol optical thickness ****/
    if (input.aer.scale_tau >= 0.0) 
      for (lc=0; lc<output->aer.nlev-1; lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++)
	  output->aer.optprop.dtau[iv][lc] = output->aer.optprop.dtau[iv][lc] * input.aer.scale_tau;


    /**** Set aerosol optical thickness according to Angstrom parameters ****/
    if (input.aer.beta >= 0.0) {
      status = scale_aer_angstrom (output->aer.optprop.dtau, output->wl.lambda_t, output->wl.nlambda_t, 
				   output->aer.nlev-1, input.aer.alpha, input.aer.beta,
				   output->aer.zd, output->atm.altitude);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by scale_aer_angstrom()\n", status);
	return status;

      }
    }

    /* Set the aerosol optical thickness to a user-defined value at 550nm */
    if (input.aer.tau550 >= 0.0) {
      lambda = calloc (output->wl.nlambda_t, sizeof(double));
      tautot = calloc (output->wl.nlambda_t, sizeof(double));

      /* determine level number of 'altitude' level */
      for (ialt=0; ialt<output->aer.nlev; ialt++)
	if (output->atm.altitude == output->aer.zd[ialt])
	  break;
      
      if (ialt == output->aer.nlev) {
	fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid\n", output->atm.altitude);
	return -1;
      }

      for (iv=0; iv<output->wl.nlambda_t; iv++) {
	lambda[iv] = (double) output->wl.lambda_t[iv];

	for (lc=0; lc<=ialt-1; lc++)
	  tautot[iv] += output->aer.optprop.dtau[iv][lc];
      }
      
      status = linear_coeffc (lambda, tautot, output->wl.nlambda_t, &a0, &a1, &a2, &a3);
      if (status!=0) {
	fprintf (stderr, "Error %d calculating interpolation coefficients for aerosol optical thickness.\n", 
		 status);
	return status;
      }
      
      status = calc_splined_value (550.0, &tau550, lambda, output->wl.nlambda_t, a0, a1, a2, a3);
      if (status!=0) {
	  fprintf (stderr, "Error %d interpolating aerosol optical thickness to 550nm.\n", 
		   status);
	  fprintf (stderr, "In the current implementation of xx_set_tau550, it is required\n");
	  fprintf (stderr, "that the internal wavelength range includes 550nm.\n");
	  fprintf (stderr, "Hope to change that in future!\n");
	  return status;
      }
    
      if (tau550>0.0) {
	  scale_factor = input.aer.tau550 / tau550;
	  
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	      for (lc=0; lc<output->aer.nlev-1; lc++)
		  output->aer.optprop.dtau[iv][lc] *= scale_factor;
      }
      

      free(a0); free(a1); free(a2); free(a3);
      free(lambda); free(tautot);
    }
  }
  else {
    /* No aerosols included in this run */

    status = calloc_aer_out (&(output->aer), output->wl.nlambda_t, 2);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for aer_out-struct\n", status);
      return status;
    }

    output->aer.zd[0] = output->atm.zd[0];
    output->aer.zd[1] = output->atm.zd[output->atm.nlev-1];
    output->aer.nlev = 2; 
  }
  
  /* add the aerosol levels to the common profile */
  set_common_z_grid (output->aer.zd, output->aer.nlev, 
		     &output->atm.zd_common, &output->atm.nlev_common);

  return 0;
}



/**************************************************************/
/* Read aerosol profile.                                      */
/**************************************************************/

static int read_aerosol_profile (char *filename, int nlyr, int nlambda_t, 
				 float *zd_common, float **dtau, int scale) 
{
  int rows=0;
  int i=0, j=0, lc=0;
  int status = 0;
  float *aero_zd=NULL, *aero_tau=NULL, *aero_tau_common=NULL;


  /* read altitude and optical depth from aerosol file */
  status = read_2c_file_float (filename, 
			       &aero_zd, &aero_tau, &rows);

  if (status!=0) {
    fprintf (stderr, "Error %d reading aerosol file %s\n", status, filename);
    return status;
  }

  aero_tau_common = calloc (nlyr+1, sizeof (float));
  
  status = regrid_float (aero_zd, aero_tau,  rows,
			 zd_common, aero_tau_common, nlyr+1, scale);
  
  if (status!=0) {
    fprintf (stderr, "Error regridding %s to common aerosol grid\n", filename);
    for (lc=0; lc<nlyr+1; lc++)
      fprintf (stderr, "%d %f\n", lc, aero_tau_common[lc]);
    

    return status;
  }

  /* copy optical depth to final destination dtau[][] */
  for (i=0; i<nlambda_t; i++)
    for (j=0; j<nlyr; j++)
      dtau[i][j] = aero_tau_common[j+1];

  /* free memory */
  free (aero_zd); free (aero_tau);
  free (aero_tau_common);

  return 0;
}



/**************************************************************/
/* Read moments of the aerosol phase function; memory for     */
/* float *moment is allocated automatically.                  */
/**************************************************************/

static int read_aerosol_moments (char *filename, 
	                         float **moment, int *nmom) 
{
  int i=0;
  int status = 0;
  double *aero_mom=NULL;
  
  /* read moments of the aerosol phase function from file */
  status = read_1c_file (filename, &aero_mom, nmom);
  
  if (status!=0) {
    switch(status)  {
    case ASCIIFILE_NOT_FOUND: 
      fprintf (stderr, "Aerosol tau file %s not found\n", filename);
      return status;
      break;
    case ASCII_NO_MEMORY: 
      fprintf (stderr, "Not enough memory for reading aerosol moments file %s\n",
	       filename);
      return status;
      break;
    default:
      fprintf (stderr, "Error %d reading aerosol file %s\n",  
	       status, filename);
      return status;
      break;
    }
  }


  /* normalize and copy moments to final destination */

  if (aero_mom[0]<=0) {
    fprintf (stderr, "Error, first moment cannot be zero or negative\n");
    return -1;
  }


  *moment = calloc (*nmom, sizeof(float));
  for (i=0;i<*nmom;i++)
    (*moment)[i] = aero_mom[i] / aero_mom[0];

  /* free memory */
  free (aero_mom);

  return 0;
}




/**************************************************************/
/* Scale the total aerosol optical thickness with Angstrom    */
/* parameters alpha and beta.                                 */
/**************************************************************/

static int scale_aer_angstrom (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			       float alpha, float beta, float *zd, float altitude)
{
  int lc=0, iv=0, ialt=0;
  float fact=0, depth=0, ext=0;
  
  /* determine level number of 'altitude' level */
  for (ialt=0; ialt<=nlyr; ialt++)
    if (altitude == zd[ialt])
      break;
  
  if (ialt == nlyr+1) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid\n", altitude);
    return -1;
  }


  /* Function Body */
  for (iv=0; iv<nlambda_t; iv++) {
    
    ext = 0.0;
    for (lc=0; lc<=ialt-1; lc++)
      ext += dtau[iv][lc];
    
    depth = beta * pow(lambda_t[iv]/1000.0, -alpha);
    fact = depth / ext;
    
    for (lc=0; lc<nlyr; lc++)
      dtau[iv][lc] *= fact;
  }
  
  return 0;
} 



/**************************************************************/
/* Scale the total aerosol optical thickness to a given value */
/**************************************************************/

static int scale_aer_tau (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			  float tau, float *zd, float altitude)
{
  int lc=0, iv=0, ialt=0;
  float ext=0, fact=0;
 
  /* determine level number of 'altitude' level */
  for (ialt=0; ialt<=nlyr; ialt++)
    if (altitude == zd[ialt])
      break;
  
  if (ialt == nlyr+1) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid\n", altitude);
    return -1;
  }

  for (iv=0; iv<nlambda_t; iv++) {
 
    ext = 0.0;
    for (lc=0; lc<=ialt-1; lc++)
      ext += dtau[iv][lc];
    
    fact = tau / ext;
    
    for (lc=0; lc<nlyr; lc++)
      dtau[iv][lc] *= fact;
  }
  
  return 0;
} 




int aerosol_set_visibility (aer_inp_struct inp, char *path, float *visib, int verbose)
{
  int status=0;

  char shettle_filename[FILENAME_MAX] = "";

  double *t0=NULL, *t1=NULL, *t2=NULL, *t3=NULL;
  double *v0=NULL, *v1=NULL, *v2=NULL, *v3=NULL;
  double *shettle_vis=NULL, *shettle_tau550=NULL;
  int shettle_n=0;
  double tau550=0, visibility=0;

  double first=0, last=0;


  /* set visibility */
  *visib = inp.visibility;
    
  /* we may need to correct the visibility if the user fixed the optical */
  /* thickness and did not adjust the visibility properly;               */
  if (inp.set_tau   >= 0.0 || 
      inp.scale_tau >= 0.0 || 
      inp.beta      >= 0.0 ||
      inp.tau550    >= 0.0)  {

    if (verbose)
      fprintf (stderr, " ... need to adjust visibility because user changed the optical thickness:\n");


      /* Read table of visibility vs. integrated optical thickness at 550nm; */
      /* the extinction profile is determined by aerosol_season and          */
      /* aerosol_vulcan whereas aerosol_haze and relative humidity only      */
      /* affect asymmetry parameter and single scattering albedo.            */

    sprintf (shettle_filename, "%s/aerosol/shettle/tau550.%1d.%1d", 
	     path, inp.vulcan, inp.seasn);

    status = read_2c_file (shettle_filename, 
			   &shettle_vis, &shettle_tau550, &shettle_n);
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, shettle_filename);
      return status;
    }

    /* now determine the user-defined optical thickness at 550nm */

    if (inp.set_tau >= 0.0) {
      tau550 = inp.set_tau;

      if (verbose)
	fprintf (stderr, " ...   aerosol_set_tau sets the optical thickness at 550nm to %f\n", 
		 tau550);
    }
      
    if (inp.scale_tau >= 0.0) {

      /* determine lower and upper visibility limits */
      first = shettle_vis[0];
      last  = shettle_vis[shettle_n-1];
      sort_double (&first, &last);

      if (*visib >= first && *visib <= last) {
	/* calculate coefficients for tau550(vis) */
	status = slinear_coeffc (shettle_vis, shettle_tau550, shettle_n, &t0, &t1, &t2, &t3);
	if (status!=0) {
	  fprintf (stderr, "Error %d calculating interpolation coefficients for %s\n", 
		   status, shettle_filename);
	  return status;
	}
	
	status = calc_splined_value (*visib, &tau550, shettle_vis, shettle_n, t0, t1, t2, t3);
	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating optical thickness at visibility %f\n", 
		   status, *visib);
	  return status;
	}

	free(t0); free(t1); free(t2); free(t3);
      }
      else {
	if (*visib <= first)
	  tau550 = shettle_tau550[0];

	if (*visib >= last)
	  tau550 = shettle_tau550[shettle_n-1];
      }

      tau550 *= inp.scale_tau;

      if (verbose)
	fprintf (stderr, " ...   aerosol_scale_tau sets the optical thickness at 550nm to %f\n", 
		 tau550);
    }

    if (inp.beta >= 0.0) {
      tau550 = inp.beta * pow(0.55, -inp.alpha);
	
      if (verbose)
	fprintf (stderr, " ...   aerosol_angstrom sets the optical thickness at 550nm to %f\n", 
		 tau550);
    }

      
    if (inp.tau550 >= 0.0) {
      tau550 = inp.tau550;

      if (verbose)
	fprintf (stderr, " ...   aerosol_set_tau550 sets the optical thickness at 550nm to %f\n", 
		 tau550);
    }



    /* determine lower and upper visibility limits */
    first = shettle_tau550[0];
    last  = shettle_tau550[shettle_n-1];
    sort_double (&first, &last);


    /* now determine the visibility which corresponds to tau550 */
    if (tau550 >= first && tau550 <= last) {

      /* calculate linear interpolation coefficients for vis(tau550) */
      status = slinear_coeffc (shettle_tau550, shettle_vis, shettle_n, &v0, &v1, &v2, &v3);
      if (status!=0) {
	fprintf (stderr, "Error %d calculating interpolation coefficients for %s\n", 
		 status, shettle_filename);
	return status;
      }

      status = calc_splined_value (tau550, &visibility, shettle_tau550, shettle_n, v0, v1, v2, v3);

      if (status!=0) {
	fprintf (stderr, "Error %d interpolating visibility at optical thickness %f\n", 
		 status, tau550);
	return status;
      }
	
      free(v0); free(v1); free(v2); free(v3);
    }
    else {
      if (tau550 <= first)
	visibility = shettle_vis[0];
	
      if (tau550 >= last)
	visibility = shettle_vis[shettle_n-1];
    }
      
    *visib = visibility;

    if (verbose)
      fprintf (stderr, " ...   changing visibility to %f km\n", *visib);
  }

  return 0;
}



int calloc_aer_out (aer_out_struct *out, int nlambda, int nlev) 
{
  int  status=0;

  out->nlev    = nlev;
  out->zd      = (float *) calloc (nlev, sizeof (float));
  out->aer_pro = (float *) calloc (nlev, sizeof (float));
		     
  status = calloc_optprop (&(out->optprop), nlambda, nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for aerosol optical properties\n", status);
    return status;
  }

  return 0;

}

int cp_aer_out (aer_out_struct *source, aer_out_struct *target, int nlambda, int alloc_moment, int quiet) 
{
  int lc=0, status=0;
  
  for (lc=0; lc<target->nlev; lc++) {
    target->zd[lc] = source->zd[lc];
    target->aer_pro[lc] = source->aer_pro[lc];
  }
  
  status = cp_optprop (&(target->optprop), &(source->optprop), source->nlev, alloc_moment, quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying aerosol optical properties\n", status);
    return status;
  }

  return 0;
}


int free_aer_out (aer_out_struct *out, int nlambda) 
{
  free (out->zd);
  free (out->aer_pro);

  free_optprop (&(out->optprop));
		     
  return 0;
}



