/*--------------------------------------------------------------------
 * $Id: solve_rte.c,v 1.79 2004/08/04 23:00:58 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <math.h>
#include <string.h>
#include <time.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/numeric.h"
#include "../src_c/fortran_and_c.h"
#include "../src_c/radiative.h"
#include "cloud.h"


#if HAVE_MYSTIC
#include "../src_c/mystic.h"
#endif

#if HAVE_SOS
#include "../src_c/sos.h"
#endif
#include "../src_f/f77-uscore.h"
#include "../src_f/solver.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif



/* internal structures */
typedef struct {
  char    *deltam;
  char    *ground_type;
  char    polscat[15];
  pol_complex ground_index;
  double  albedo;
  double  btemp;
  double  flux;
  double  *gas_extinct;
  double  *height;
  double  mu;
  double  sky_temp;
  double  *temperatures;
  double  wavelength;
  int     *outlevels;
  int     nummu; /* Number of quadrature angles (per hemisphere). */
} polradtran_input;


typedef struct {
  char header[127]; /* A 127- (or less) character header for prints         */

  float accur;      /* Convergence criterion for azimuthal series.          */

  float fbeam;      /* Intensity of incident parallel beam at top boundary. */
                    /*  [same units as PLKAVG (default W/sq m) if thermal   */
                    /*  sources active, otherwise arbitrary units].         */

  float fisot;      /* Intensity of top-boundary isotropic illumination.    */
                    /*  [same units as PLKAVG (default W/sq m) if thermal   */
                    /*  sources active, otherwise arbitrary units].         */
                    /*  Corresponding incident flux is pi (3.14159...)      */
                    /*  times FISOT.                                        */

  float *hl;        /* K = 0 to NSTR.  Coefficients in Legendre-            */
                    /* polynomial expansion of bottom bidirectional         */
                    /* reflectivity                                         */

  int   planck;     /* TRUE,  include thermal emission                      */
                    /* FALSE, ignore thermal emission (saves computer time) */

  float btemp;      /* Temperature of bottom boundary (K)                   */
                    /* (bottom emissivity is calculated from ALBEDO or HL,  */
                    /* so it need not be specified).                        */
                    /* Needed only if PLANK is TRUE.                        */

  float ttemp;      /* Temperature of top boundary (K)                      */
                    /* Needed only if PLANK is TRUE.                        */

  float temis;      /* Emissivity of top boundary.                          */
                    /* Needed only if PLANK is TRUE.                        */

  float *utau;
  float umu0 ;

  int ierror_d[47];
  int ierror_s[33];
  int ierror_t[22];

  int prndis[7];
  int prndis2[5];
  int prntwo[2];

  int ibcnd;        /* 0 : General case.                                    */
                    /* 1 : Return only albedo and transmissivity of the     */
                    /*     entire medium vs. incident beam angle.           */

  int lamber;       /* TRUE, isotropically reflecting bottom boundary.      */
                    /* FALSE, bidirectionally reflecting bottom boundary.   */

  int onlyfl;       /* TRUE, return fluxes, flux divergences, and mean      */
		    /*       intensities.                                   */
                    /* FALSE, return fluxes, flux divergences, mean         */
                    /*        intensities, azimuthally averaged intensities */
                    /*        (at the user angles) AND intensities.         */


  int quiet;
  int usrang;
  int usrtau;

  /* sdisort-specific variables */ 
  int nil;    
  int newgeo; 
  int spher;

  /* PolRadtran-specific variables */ 
  polradtran_input pol;

} rte_input;



typedef struct {
  float *dfdt;
  float *flup;
  float *rfldir;
  float *rfldn;
  float *uavgso;
  float *uavgdn;
  float *uavgup;
  float *uavg;
  float *heat;
  float **u0u;
  float ***uu;
  	 
  float ***rfldir3d;
  float ***rfldn3d;
  float ***flup3d;
  float ***abs3d;
  float ***radiance3d;

  float *albmed;
  float *trnmed;

  double **polradtran_down_flux;
  double **polradtran_up_flux;
  double ****polradtran_down_rad_q; /* _q indicates radiances at quadrature angels. */
  double ****polradtran_up_rad_q;
  double ****polradtran_down_rad;
  double ****polradtran_up_rad;
  double *polradtran_mu_values;
} rte_output;




/* prototypes of internal functions */
static int reverse_profiles (input_struct input, output_struct *output);
static rte_output *calloc_rte_output (input_struct input);
static int reset_rte_output (rte_output **rte, input_struct input);
static void free_rte_output (rte_output *result, input_struct input);
static int add_rte_output (rte_output *rte, rte_output *add, double factor, input_struct input);
static int init_rte_input (rte_input *rte, input_struct input, output_struct *output);
static int setup_and_call_solver (input_struct input, output_struct *output, 
				  rte_input *rte_in, rte_output *rte_out, 
				  int iv, int ib, int mc_loaddata);
static int call_solver (input_struct input, output_struct *output,
			rte_input *rte_in, int iv, rte_output *rte_out,
			int mc_loaddata, int verbose);
static void fourier2azimuth (double**** down_rad_rt3, double**** up_rad_rt3,
			     double**** down_rad, double**** up_rad, 
			     int nzout, int aziorder,
			     int nstr, int numu, int nstokes,
			     int nphi, float* phi);

static int cp_3d_to_rte_output (rte_output *rte, 
				float *rfldir3d, float *rfldn3d, float *flup3d, float *radiance3d,
				int Nz, int Nx, int Ny);

static int cp_abs_to_rte_output (rte_output *rte, 
				 float *abs3d,
				 int Nz, int Nx, int Ny);

static int calc_spectral_heating (float* heat, float* dfdt,
			      float* zd, float* dens_air, float* utau, input_struct input );



/*********************************************************************/
/* Main function. Loop over wavelengths or wavelength bands,         */
/* correlated-k quadrature points, and independent pixels.           */
/*********************************************************************/

int solve_rte (input_struct input, output_struct *output)
{
  static int first=1;
  int status=0;
  int ic=0, is=0, js=0, ks=0, iu=0, iv=0, j=0, lu=0;
  int iq=0;

  int mc_loaddata=1;

  rte_input  *rte_in      = NULL;
  rte_output *rte_out     = NULL;
  rte_output *rte_outband = NULL;

  if (first) {
    first=0;

    rte_in      = calloc(1, sizeof(rte_input));
    rte_out     = calloc_rte_output(input);
    rte_outband = calloc_rte_output(input);

    /* initialize RTE input */
    status = init_rte_input (rte_in, input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d initializing rte input\n", status);
      return status;
    }
    
    /* allocate memory for the output result structures */
    status = setup_result (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for the model output\n", status);
      return status;
    }
  }

  /************************/
  /* loop over wavelength */
  /************************/
  for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {

    /* solar zenith angle at this wavelength */
    rte_in->umu0 = cos(output->atm.sza_t[iv]*PI/180.0);
    
    /********************************/
    /* loop over independent pixels */
    /********************************/

    for (ic=0; ic<output->nipa; ic++) {
      /* copy pixel number ic to 1D cloud data */

      if (input.wcipa) {

	/* copy everything except the single scattering properties */
	status = cp_cld_out (&(output->wc), output->wcipa[ic], 0, 0, input.quiet);
	if (status!=0) {
	  fprintf (stderr, "Error %d copying output->wcipa[ic] to output->wc\n", status);
	  return status;
	}

	switch (input.wc.properties) {
	case WCPROP_MIE:
	case WCPROP_FILE:
	  
	  /* recalculate the optical properties for the */
	  /* new microphysics profile                   */
	  status = ssprop2optprop (&(output->wc), iv);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d calculating water cloud optical properties\n", status);
	    return status;
	  }
	  
	  /* overwrite these properties with user-defined optical thickness, ssa, etc */
	  status = apply_user_defined_properties_to_cld (input.wc, 
							 output->atm, 
							 output->wl, 
							 output->atm.altitude,
							 &(output->wc));
	  if (status!=0) {
	    fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
		     status);
	    return status;
	  }
	  
	case WCPROP_NONE:
	case WCPROP_HU:
	case WCPROP_EXPLICIT:
	  break;
	  
	default:
	  fprintf (stderr, "Error, unknown water cloud property %d. This is not\n", input.wc.properties);
	  fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	  
	  return -1;
	  break;
	}
      }
      

      if (input.icipa) {

	/* copy everything except the single scattering properties */
	status = cp_cld_out (&(output->ic), output->icipa[ic], 0, 0, input.quiet);
	if (status!=0) {
	  fprintf (stderr, "Error %d copying output->icipa[ic] to output->ic\n", status);
	  return status;
	}

	if (input.ic.properties == ICPROP_MIE || input.ic.properties == ICPROP_FILE) {

	  /* recalculate the optical properties for the */
	  /* new microphysics profile                    */
	  status = ssprop2optprop (&(output->ic), iv);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d calculating ice cloud optical properties\n", status);
	    return status;
	  }

	  /* overwrite these properties with user-defined optical thickness, ssa, etc */
	  status = apply_user_defined_properties_to_cld (input.ic, 
							 output->atm, 
							 output->wl, 
							 output->atm.altitude,
							 &(output->ic));
	  if (status!=0) {
	    fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
		     status);
	    return status;
	  }
	}
      }
      

      /*
      if (input.verbose) {
	fprintf (stderr, "*** wavelength: iv = %d, %f nm\n", iv, output->wl.lambda_t[iv]);
	fprintf (stderr, "    atm.nmom + 1 = %d phase function moments\n", output->atm.nmom+1);
	fprintf (stderr, " ---------------------------------------------------------------------------\n");
	fprintf (stderr, "   lc |    z[km] |      aerosol      |    water cloud    |    ice cloud     \n");
	fprintf (stderr, "      |          |        dtau  nmom |        dtau  nmom |        dtau  nmom\n");
	fprintf (stderr, " ---------------------------------------------------------------------------\n");
	for (lu=0; lu<output->atm.nlyr; lu++)
	  fprintf (stderr, "%5d | %8.2f | %11.6f %5d | %11.6f %5d | %11.6f %5d\n", 
		   lu, output->atm.zd[lu+1],
		   output->aer.dtau[iv][lu], output->aer.nmom[iv][lu],
		   output->wc.optprop.dtau [iv][lu], output->wc.optprop.nmom[iv][lu],
		   output->ic.optprop.dtau [iv][lu], output->ic.optprop.nmom[iv][lu]);
	fprintf (stderr, " ---------------------------------------------------------------------------\n");
      }    
      */

      /* need to load data during first call of mystic() for each band/wavelength */
      mc_loaddata=1;  

      /* reset band integral */
      reset_rte_output(&rte_outband, input);

      /*******************************/
      /* loop over quadrature points */
      /*******************************/
      for (iq=0; iq<output->atm.nq_t[iv]; iq++) {

	/* set number of photons for this band */
	output->mc_photons = (int) (output->mc_photons_t[iv][iq] * (double) input.rte.mc.photons + 0.5);
	
	/* run at least MIN_MCPHOTONS for each band */
	if (output->mc_photons < MIN_MCPHOTONS)
	  output->mc_photons = MIN_MCPHOTONS;
	
	/* setup optical properties and call the RTE solver */
	status = setup_and_call_solver (input, output, rte_in, rte_out, iv, iq, mc_loaddata);
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by setup_and_call_solver()\n", status);
	  return status;
	}
	
	/* data need to be loaded only once per iv */
	mc_loaddata=0;
	
	/* add result for the current quadrature point considering quadrature weight */
	add_rte_output (rte_outband, rte_out, output->atm.wght_t[iv][iq], input);
	if ( input.processing ==  PROCESS_HEATING ) {
	  calc_spectral_heating (rte_outband->heat, rte_out->dfdt,  input.atm.zout_sur, 
				 output->atm.microphys.dens_air, rte_in->utau, input);
	}
      }


      /************************************************/
      /* add result for the current independent pixel */
      /************************************************/

      if (input.rte.solver == POLRADTRAN ) {
	for(lu=0; lu<input.atm.nzout; lu++) {
	  for(is=0; is<input.rte.pol_nstokes; is++) {

	    output->rfldir_t [lu][iv] += output->ipaweight[ic] * rte_outband->rfldir [lu];


	    output->up_flux_t[lu][is][iv]   += 
	      output->ipaweight[ic] * rte_outband->polradtran_up_flux[lu][is]; 
	    output->down_flux_t[lu][is][iv] += 
	      output->ipaweight[ic] * rte_outband->polradtran_down_flux[lu][is];

	    for (j=0;j<input.rte.nphi;j++) {
	      for (iu=0;iu<input.rte.nstr/2;iu++) {
		output->down_rad_t[lu][j][iu][is][iv] += 
		  output->ipaweight[ic] * rte_outband->polradtran_down_rad[lu][j][iu][is];
		output->up_rad_t[lu][j][iu][is][iv] += 
		  output->ipaweight[ic] * rte_outband->polradtran_up_rad[lu][j][iu][is];
	      }
	    }
	  }
	}
      }
      else {
	for(lu=0; lu<input.atm.nzout; lu++) {
	  output->rfldir_t [lu][iv] += output->ipaweight[ic] * rte_outband->rfldir [lu];
	  output->rfldn_t  [lu][iv] += output->ipaweight[ic] * rte_outband->rfldn  [lu];
	  output->flup_t   [lu][iv] += output->ipaweight[ic] * rte_outband->flup   [lu];
	  output->uavg_t   [lu][iv] += output->ipaweight[ic] * rte_outband->uavg   [lu];
	  output->uavgdn_t [lu][iv] += output->ipaweight[ic] * rte_outband->uavgdn [lu];
	  output->uavgso_t [lu][iv] += output->ipaweight[ic] * rte_outband->uavgso [lu];
	  output->uavgup_t [lu][iv] += output->ipaweight[ic] * rte_outband->uavgup [lu];
	  
	  /* intensities */
	  for (iu=0;iu<input.rte.numu;iu++) {
	    output->u0u_t[lu][iu][iv] += output->ipaweight[ic]*rte_outband->u0u[lu][iu];
	    
	    for (j=0;j<input.rte.nphi;j++)
	      output->uu_t[lu][j][iu][iv] += output->ipaweight[ic]*rte_outband->uu[j][lu][iu];
	  }
	  
	  /* 3D fields */
	  if (input.rte.mc.passback3D) {
	    for (is=0; is<input.rte.mc.Nx_sample; is++) 
	      for (js=0; js<input.rte.mc.Ny_sample; js++) {
		
		output->rfldir3d_t   [lu][is][js][iv] += 
		  output->ipaweight[ic] * rte_outband->rfldir3d   [lu][is][js];

		output->rfldn3d_t    [lu][is][js][iv] += 
		  output->ipaweight[ic] * rte_outband->rfldn3d    [lu][is][js];

		output->flup3d_t     [lu][is][js][iv] += 
		  output->ipaweight[ic] * rte_outband->flup3d     [lu][is][js];

		output->radiance3d_t [lu][is][js][iv] += 
		  output->ipaweight[ic] * rte_outband->radiance3d [lu][is][js];
	      }
	  } 
	}

	/* 3D absorption */
	if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE) 
	  for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	    for (is=0; is<input.rte.mc.Nx_cloud; is++)
	      for (js=0; js<input.rte.mc.Ny_cloud; js++) 
		output->abs3d_t [ks][is][js][iv] += 
		  output->ipaweight[ic] * rte_outband->abs3d [ks][is][js];
      }
    }
  }
  



  /* free temporary memory */

  /* ??? why is that commented out ??? 
  free(rte_in);
  free_rte_output (rte_out, input);
  free_rte_output (rte_outband, input);
  */

  if (input.rte.solver == POLRADTRAN) {
    free (rte_in->pol.height);
    free (rte_in->pol.temperatures);
    free (rte_in->pol.gas_extinct);
    free (rte_in->pol.outlevels);
  }


  return 0;
}



int setup_result (input_struct input, output_struct *output)
{
  int status=0;
  int nlambda=0;

  if ((status = ASCII_calloc_float (&output->flup_t,   input.atm.nzout, output->wl.nlambda_t)) != 0)
    return status;

  if ((status = ASCII_calloc_float (&output->rfldir_t, input.atm.nzout, output->wl.nlambda_t)) != 0)
    return status;

  if ((status = ASCII_calloc_float (&output->rfldn_t,  input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;
  
  if ((status = ASCII_calloc_float (&output->uavg_t,   input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ((status = ASCII_calloc_float (&output->uavgdn_t, input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ((status = ASCII_calloc_float (&output->uavgso_t, input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ((status = ASCII_calloc_float (&output->uavgup_t, input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ( input.processing ==  PROCESS_HEATING) 
    output->heat = (float *) calloc (input.atm.nzout, sizeof(float));
  

  if (input.rte.maxumu>0)
    if ((status = ASCII_calloc_float_3D (&output->u0u_t, input.atm.nzout, input.rte.maxumu,
					 output->wl.nlambda_t)) != 0)
      return status;
  
  if (input.rte.maxumu>0 && input.rte.maxphi>0)
    if ((status = ASCII_calloc_float_4D (&output->uu_t,  input.atm.nzout, input.rte.maxphi, 
				       input.rte.maxumu, output->wl.nlambda_t)) !=0)
      return status;
  

  if (input.rte.mc.passback3D) {

    if ((status = ASCII_calloc_float_4D (&output->rfldir3d_t, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->rfldn3d_t, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;

    if ((status = ASCII_calloc_float_4D (&output->flup3d_t, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->radiance3d_t, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    


    /* 3d absorption */
    if (input.rte.mc.absorption!=MCABS_NONE)
      if ((status = ASCII_calloc_float_4D (&output->abs3d_t, 
					   input.rte.mc.Nz_cloud, input.rte.mc.Nx_cloud,
					   input.rte.mc.Ny_cloud, output->wl.nlambda_t)) != 0)
	return status;
    
      
  }

  
  /* need to allocate enough memory for both output->wl.nlambda_s */
  /* and output->wl.nlambda_h, hence using whichever is larger    */
  nlambda = (output->wl.nlambda_h > output->wl.nlambda_s ?
	     output->wl.nlambda_h : output->wl.nlambda_s);

  if ((status = ASCII_calloc_float (&output->flup,   input.atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->rfldn,  input.atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->rfldir, input.atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavg,   input.atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgdn, input.atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgso, input.atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgup, input.atm.nzout, nlambda)) != 0) 
    return status;
  
  if ((status = ASCII_calloc_float_3D (&output->down_flux, input.atm.nzout, input.rte.pol_nstokes, 
				       nlambda)) != 0) 
    return status;
    
  if ((status = ASCII_calloc_float_3D (&output->down_flux_t,   input.atm.nzout, input.rte.pol_nstokes, 
				       output->wl.nlambda_t)) !=0 )
    return status;
    
  if ((status = ASCII_calloc_float_3D (&output->up_flux, input.atm.nzout, input.rte.pol_nstokes, 
				       nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float_3D (&output->up_flux_t,   input.atm.nzout, input.rte.pol_nstokes, 
				       output->wl.nlambda_t)) !=0 )
    return status;
    
  if (input.rte.nphi>0) {
    if ((status = ASCII_calloc_float_5D (&output->down_rad, input.atm.nzout, input.rte.nphi,
					 input.rte.nstr/2, input.rte.pol_nstokes, 
					 nlambda)) !=0 )
      return status;

    if ((status = ASCII_calloc_float_5D (&output->down_rad_t, input.atm.nzout, input.rte.nphi,
					 input.rte.nstr/2, input.rte.pol_nstokes, 
					 output->wl.nlambda_t)) !=0 )
      return status;
    
    if ((status = ASCII_calloc_float_5D (&output->up_rad, input.atm.nzout, input.rte.nphi,
					 input.rte.nstr/2, input.rte.pol_nstokes, 
					 nlambda)) !=0 )
      return status;

    if ((status = ASCII_calloc_float_5D (&output->up_rad_t, input.atm.nzout, input.rte.nphi,
					 input.rte.nstr/2, input.rte.pol_nstokes, 
					 output->wl.nlambda_t)) !=0 )
      return status;
  }

  if (input.rte.maxumu>0)
    if ((status = ASCII_calloc_float_3D (&output->u0u, input.atm.nzout, input.rte.maxumu, 
					 nlambda)) != 0) 
      return status;

  if (input.rte.maxumu>0 && input.rte.maxphi>0)
    if ((status = ASCII_calloc_float_4D (&output->uu, input.atm.nzout, input.rte.maxphi, 
					 input.rte.maxumu, nlambda)) != 0) 
      return status;

  if (input.rte.mc.passback3D) {
    if ((status = ASCII_calloc_float_4D (&output->rfldir3d, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->rfldn3d, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;

    if ((status = ASCII_calloc_float_4D (&output->flup3d, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;

    if ((status = ASCII_calloc_float_4D (&output->radiance3d, 
					 input.atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;
    

    /* 3D absorption */
    if (input.rte.mc.absorption!=MCABS_NONE)
      if ((status = ASCII_calloc_float_4D (&output->abs3d, 
					   input.rte.mc.Nz_cloud, input.rte.mc.Nx_cloud,
					   input.rte.mc.Ny_cloud, nlambda)) != 0)
	return status;

  }


  if ((output->sza_h = (float *) calloc (output->wl.nlambda_h, sizeof(float))) == NULL)
    return -1;
  
  return 0;
}

static int reverse_profiles (input_struct input, output_struct *output)
{
  int lu=0, iu=0;
  double tmp=0;

  for (lu=0; lu<output->atm.nlyr/2; lu++) {
    
    /* reverse profile of optical depth dtauc */
    tmp               = output->dtauc[output->atm.nlyr-lu-1];
    output->dtauc[output->atm.nlyr-lu-1] = output->dtauc[lu];
    output->dtauc[lu] = tmp;
    
    /* reverse profile of single scattering albedo ssalb */
    tmp               = output->ssalb[output->atm.nlyr-lu-1];
    output->ssalb[output->atm.nlyr-lu-1] = output->ssalb[lu];
    output->ssalb[lu] = tmp;
    
    /* reverse profile of phase function pmom */
    for (iu=0;iu<=input.rte.nstr;iu++) {
      tmp                  =  output->pmom[output->atm.nlyr-lu-1][iu];
      output->pmom[output->atm.nlyr-lu-1][iu] = output->pmom[lu][iu];
      output->pmom[lu][iu] = tmp;
    }
  }
  return 0;
}


/* allocate temporary memory for the RTE solvers */
static rte_output *calloc_rte_output (input_struct input)
{
  int status=0;


  rte_output *result = calloc(1, sizeof(rte_output));
  
  result->albmed = (float *) calloc (input.rte.maxumu, sizeof(float));
  result->trnmed = (float *) calloc (input.rte.maxumu, sizeof(float));
	     
  result->dfdt   = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->flup   = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->rfldir = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->rfldn  = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavg   = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavgdn = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavgso = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavgup = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->heat   = (float *) calloc (input.atm.nzout,  sizeof(float));
  
  if (input.rte.maxumu>0)
    if ((status = ASCII_calloc_float    (&(result->u0u), 
					 input.atm.nzout, input.rte.maxumu)) != 0) 
      return NULL;
  
  if (input.rte.maxphi>0 && input.rte.maxumu>0)
    if ((status = ASCII_calloc_float_3D (&(result->uu), 
					 input.rte.maxphi, input.atm.nzout, input.rte.maxumu)) != 0) 
      return NULL;
  

  /* PolRadtran-specific */
  if (input.rte.solver == POLRADTRAN) {
    result->polradtran_mu_values = (double *) calloc (input.rte.nstr/2+input.rte.numu, sizeof(double));
    
    if ((status = ASCII_calloc_double (&(result->polradtran_up_flux), 
				       input.atm.nzout, input.rte.pol_nstokes)) != 0)
      return NULL;
    
    
    if ((status = ASCII_calloc_double (&(result->polradtran_down_flux), 
				       input.atm.nzout, input.rte.pol_nstokes)) != 0)
      return NULL;


    if ((status = ASCII_calloc_double_4D (&(result->polradtran_up_rad), 
					  input.atm.nzout, input.rte.nphi,
					  input.rte.nstr/2+input.rte.numu, input.rte.pol_nstokes )) != 0)
      return NULL;

    if ((status = ASCII_calloc_double_4D (&(result->polradtran_down_rad), 
					  input.atm.nzout, input.rte.nphi,
					  input.rte.nstr/2+input.rte.numu, input.rte.pol_nstokes )) != 0)
      return NULL;


    if ((status = ASCII_calloc_double_4D (&(result->polradtran_up_rad_q), 
					  input.atm.nzout, input.rte.pol_aziorder+1,
					  input.rte.nstr/2+input.rte.numu, input.rte.pol_nstokes )) != 0)
      return NULL;

    if ((status = ASCII_calloc_double_4D (&(result->polradtran_down_rad_q), 
					  input.atm.nzout, input.rte.pol_aziorder+1,
					  input.rte.nstr/2+input.rte.numu, input.rte.pol_nstokes )) != 0)
      return NULL;

  }


  /* 3d fields */
  if (input.rte.mc.passback3D) {

    if ((status = ASCII_calloc_float_3D (&(result->rfldir3d), 
					 input.atm.nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->rfldn3d), 
					 input.atm.nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->flup3d), 
					 input.atm.nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->radiance3d), 
					 input.atm.nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	


    if (input.rte.mc.absorption!=MCABS_NONE) {

      if ((status = ASCII_calloc_float_3D (&(result->abs3d), 
					   input.rte.mc.Nz_cloud,
					   input.rte.mc.Nx_cloud,
					   input.rte.mc.Ny_cloud)) != 0)
	return NULL;
    }
  }   
  return result;
}


/* reset rte_output structure */
static int reset_rte_output (rte_output **rte, input_struct input) {
  free_rte_output(*rte, input);
  *rte = calloc_rte_output (input);

  return 0; /* if o.k. */
}

  
static int add_rte_output (rte_output *rte, rte_output *add, double factor, input_struct input)
{
  int j=0, iu=0, is=0, js=0, ks=0, lu=0;
  
  for(lu=0; lu<input.atm.nzout; lu++) {
    
    rte->rfldir [lu] += factor*add->rfldir [lu];
    rte->rfldn  [lu] += factor*add->rfldn  [lu];
    rte->flup   [lu] += factor*add->flup   [lu];
    rte->uavg   [lu] += factor*add->uavg   [lu];
    rte->uavgdn [lu] += factor*add->uavgdn [lu];
    rte->uavgso [lu] += factor*add->uavgso [lu];
    rte->uavgup [lu] += factor*add->uavgup [lu];
    rte->dfdt   [lu] += factor*add->dfdt   [lu];
    
    for (iu=0;iu<input.rte.numu;iu++) {
      rte->u0u[lu][iu] += factor*add->u0u[lu][iu];
      
      for (j=0;j<input.rte.nphi;j++)
	rte->uu[j][lu][iu] += factor*add->uu[j][lu][iu];
    }
  }

  for (iu=0;iu<input.rte.numu;iu++) {
    rte->albmed[iu] += factor*add->albmed[iu];
    rte->trnmed[iu] += factor*add->trnmed[iu];
  }


  /* PolRadtran-specific */
  if (input.rte.solver == POLRADTRAN) {
    for(lu=0; lu<input.atm.nzout; lu++) {
      for(is=0; is<input.rte.pol_nstokes; is++) {
	
	rte->polradtran_up_flux[lu][is]   += factor*add->polradtran_up_flux[lu][is];
	rte->polradtran_down_flux[lu][is] += factor*add->polradtran_down_flux[lu][is];
	
	for (j=0;j<input.rte.nphi;j++) {
	  for (iu=0;iu<input.rte.nstr/2;iu++) {
	    rte->polradtran_up_rad[lu][j][iu][is]   += factor*add->polradtran_up_rad[lu][j][iu][is];
	    rte->polradtran_down_rad[lu][j][iu][is] += factor*add->polradtran_down_rad[lu][j][iu][is];
	  }
	}
      }
    }
  }


  if (input.rte.mc.passback3D) {
    for (ks=0; ks<input.atm.nzout; ks++) 
      for (is=0; is<input.rte.mc.Nx_sample; is++) 
	for (js=0; js<input.rte.mc.Ny_sample; js++) {
	  rte->rfldir3d   [ks][is][js] += factor * add->rfldir3d   [ks][is][js];
	  rte->rfldn3d    [ks][is][js] += factor * add->rfldn3d    [ks][is][js];
	  rte->flup3d     [ks][is][js] += factor * add->flup3d     [ks][is][js];
	  rte->radiance3d [ks][is][js] += factor * add->radiance3d [ks][is][js];
	}
    
    if (input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++) 
	for (is=0; is<input.rte.mc.Nx_cloud; is++) 
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) 
	    rte->abs3d [ks][is][js] += factor * add->abs3d [ks][is][js];
  }    
    
  return 0;  /* if o.k. */
}



static int cp_3d_to_rte_output (rte_output *rte, 
				float *rfldir3d, float *rfldn3d, float *flup3d, float *radiance3d,
				int Nz, int Nx, int Ny)
{
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, rfldir3d,   rte->rfldir3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, rfldn3d,    rte->rfldn3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, flup3d,     rte->flup3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, radiance3d, rte->radiance3d);

  return 0;
}


static int cp_abs_to_rte_output (rte_output *rte, 
				 float *abs3d,
				 int Nz, int Nx, int Ny)
{

  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, abs3d, rte->abs3d);

  return 0;
}



/* free temporary memory for the RTE solvers */
static void free_rte_output (rte_output *result, input_struct input)
{

  if (result->albmed != NULL) free(result->albmed);

  if (result->trnmed != NULL) free(result->trnmed);
         
  if (result->dfdt != NULL)   free(result->dfdt);

  if (result->flup != NULL)   free(result->flup);

  if (result->rfldir != NULL) free(result->rfldir);

  if (result->rfldn != NULL)  free(result->rfldn);

  if (result->uavg != NULL)   free(result->uavg);

  if (result->uavgdn != NULL) free(result->uavgdn);

  if (result->uavgso != NULL) free(result->uavgso);

  if (result->uavgup != NULL) free(result->uavgup);

  if (result->heat != NULL)   free(result->heat);

  if (result->u0u != NULL)
    ASCII_free_float (result->u0u, input.atm.nzout);


  if (result->uu != NULL)
    ASCII_free_float_3D (result->uu, input.rte.maxphi, input.atm.nzout);


  if ( input.rte.solver == POLRADTRAN ) {

    if (result->polradtran_mu_values != NULL)
      free (result->polradtran_mu_values);
    
    if (result->polradtran_up_flux != NULL)
      ASCII_free_double(result->polradtran_up_flux,   input.atm.nzout);

    if (result->polradtran_down_flux != NULL)
      ASCII_free_double(result->polradtran_down_flux, input.atm.nzout);
    
    if (result->polradtran_up_rad != NULL)
      ASCII_free_double_4D(result->polradtran_up_rad, 
			   input.atm.nzout, input.rte.nphi, input.rte.nstr/2+input.rte.numu);
    
    if (result->polradtran_down_rad != NULL) 
      ASCII_free_double_4D(result->polradtran_down_rad, 
			   input.atm.nzout, input.rte.nphi, input.rte.nstr/2+input.rte.numu);

    if (result->polradtran_up_rad_q != NULL)
      ASCII_free_double_4D(result->polradtran_up_rad_q, 
			   input.atm.nzout, input.rte.pol_aziorder+1, input.rte.nstr/2+input.rte.numu);
    
    if (result->polradtran_down_rad_q != NULL)
      ASCII_free_double_4D(result->polradtran_down_rad_q, 
			   input.atm.nzout, input.rte.pol_aziorder+1, input.rte.nstr/2+input.rte.numu);
  }


  /* free 3d fields */
  if (input.rte.mc.passback3D) {

    if (result->rfldir3d != NULL)
      ASCII_free_float_3D(result->rfldir3d, 
			  input.atm.nzout, input.rte.mc.Nx_sample);

    if (result->rfldn3d != NULL)
      ASCII_free_float_3D(result->rfldn3d, 
			  input.atm.nzout, input.rte.mc.Nx_sample);

    if (result->flup3d != NULL)
      ASCII_free_float_3D(result->flup3d, 
			  input.atm.nzout, input.rte.mc.Nx_sample);
    
    if (result->radiance3d != NULL)
      ASCII_free_float_3D(result->radiance3d, 
			  input.atm.nzout, input.rte.mc.Nx_sample);
    
    if (input.rte.mc.absorption!=MCABS_NONE) {
      if (result->abs3d != NULL)
	ASCII_free_float_3D(result->abs3d, 
			    input.rte.mc.Nz_cloud, input.rte.mc.Nx_cloud);
      
    }
  }
  free(result);
}





/******************************************************/
/* setup optical properties, do some initializations, */
/* and call the RTE solver.                           */
/******************************************************/

static int setup_and_call_solver (input_struct input, output_struct *output, 
				  rte_input *rte_in, rte_output *rte_out, 
				  int iv, int ib, int mc_loaddata)
{
  int status=0, lu=0;
  static int first = 1;

  /* calculate optical properties from model input */
  status = optical_properties (input, output, iv, ib);
  
  if (status!=0) {
    fprintf (stderr, "Error %d returned by optical_properties()\n", 
	     status);
    return status;
  }
  

  if (input.ck_scheme==CK_SBDART) {

    /* this is a little inefficient as we need first the scattering optical */
    /* properties to calculate the absorption coefficients and then         */
    /* recalculate the optical properties; the reason is that SBDART        */
    /* requires the total scattering cross section as input, and this       */
    /* quantity is only available after the call to optical_properties      */

    /* ??? where to get the mixing ratio of N2 from ??? */
    /* ??? setting this value to -1                 ??? */
    status = sbdart_profile (output->atm.microphys.temper, output->atm.microphys.press, output->atm.zd, 
			     output->atm.microphys.dens[MOL_H2O], output->atm.microphys.dens[MOL_O3], 
			     -1.0, 
			     output->mixing_ratio[MX_O2], 
			     output->mixing_ratio[MX_CO2], 
			     output->mixing_ratio[MX_CH4], 
			     output->mixing_ratio[MX_N2O], 
			     output->mixing_ratio[MX_NO2], 
			     input.ck_o4abs,
			     output->dtauc, output->ssalb, output->atm.nlev, 
			     output->atm.sza_t[iv], output->wl.lambda_t[iv],
		             iv, output->crs_ck.profile[0]);

    if (status!=0) {
      fprintf (stderr, "Error %d returned by sbdart_profile()\n", status);
      return status;
    }


    /* allocate memory for absorption coefficient profile */
    if (first) {
      status  = ASCII_calloc_float (&output->kabs.kabs, output->atm.nlev, 
				    output->wl.nlambda_t);
      if (status!=0) { 
	fprintf (stderr, "Error %d allocating memory for output->kabs.kabs\n", status); 
	return status; 
      } 

      first=0;
    }
    
    /* set absorption coefficient */
    for (lu=0; lu<output->atm.nlyr; lu++)
      output->kabs.kabs[lu][iv]  = 
	output->crs_ck.profile[0][iv].crs[lu][ib];

    /* calculate optical properties from model input */
    status = optical_properties (input, output, iv, ib);
    
    if (status!=0) {
	fprintf (stderr, "Error %d returned by optical_properties()\n", 
		 status);
	return status;
    }
  }

  /* translate output altitudes (zout) to optical depths (utau) */
  if (input.rte.solver != MONTECARLO) {

    /* "old" version */
    /*
    status = set_out (output->atm.zd, output->dtauc, output->atm.nlyr, 
		      input.atm.zout, input.atm.nzout, rte_in->utau);

    if (status!=0) {
      fprintf (stderr, "Error %d returned by set_out()\n", status);
      return status;
    }
    */

    /* "very old" version, recycled */ 
    F77_FCN  (setout) (output->dtauc, 
		       &(output->atm.nlyr), 
		       &(input.atm.nzout), 
		       rte_in->utau,
		       output->atm.zd, 
		       output->atm.zout_sur);
    
  }


  /* reverse profiles, if required */
  if (input.rte.reverse) {
    status = reverse_profiles (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d reversing profiles\n", status);
      return status;
    }
  }
  
  
  /* call RTE solver */
  status = call_solver (input, output, rte_in, iv, rte_out, mc_loaddata, input.verbose);
  if (status!=0) {
    fprintf (stderr, "Error %d calling solver\n", status);
    return status;
  }
  
  return 0;  /* if o.k. */
}





static int call_solver (input_struct input, output_struct *output,
			rte_input *rte_in, int iv, rte_output *rte_out,
			int mc_loaddata, int verbose)
{
  int status=0;

  double start=0, end=0;

  int lu=0, is=0;

  float radius= 6370.0; /* Earth's radius in km */

  float *twostr_gg    = NULL;
  float *sdisort_beta = NULL;
  float *sdisort_sig  = NULL;
  float **sdisort_denstab = NULL;
  float *tosdisort_denstab = NULL;
  float *disort_pmom  = NULL, *disort2_pmom  = NULL, *sss_pmom  = NULL;

  int planck_tempoff = 0;

#if HAVE_POLRADTRAN
  int iu=0; 
 
  double *polradtran_down_flux = NULL;
  double *polradtran_up_flux   = NULL;
  double *polradtran_down_rad  = NULL;
  double *polradtran_up_rad    = NULL;
#endif


  /* temporary Fortran arrays */
  float *disort_u0u = NULL, *disort_uu = NULL;
  float *tzs_u0u = NULL, *tzs_uu = NULL;
  float *sss_u0u = NULL, *sss_uu = NULL;

#if HAVE_MYSTIC
  int il=0, nmom=0;
  int source=0;
  int thermal_photons=0;
  int nres3d=0, nabs3d=0;
  char tempstr[10]="";

  /* temporary MC output (for thermal MC calculations, two calls */
  /* to mc() are required, one for the surface and one for the   */
  /* atmospheric contribution                                    */
  rte_output *tmp_out=NULL;

  float *tmp_rfldir3d=NULL, *tmp_rfldn3d=NULL, *tmp_flup3d=NULL, *tmp_abs3d=NULL;
  float *tmp_radiance3d=NULL;
 
  /* phase function file */
  char momcld_filename[FILENAME_MAX]="";

  /* rpv arrays */
  float *rpv_rho0=NULL;
  float *rpv_k=NULL;
  float *rpv_theta=NULL;

  /* create name for MYSTIC cloud phase function file */
  if (input.rte.solver == MONTECARLO) {
    switch (input.rte.mc.momcld) {
    case MOMCLD_FILE:
      /*
      strcpy (momcld_filename, input.rte.mc.filename[FN_MC_MOMCLD]);
      break;
      */
      
    case MOMCLD_WATER:
    case MOMCLD_ICE:
      strcpy (momcld_filename, input.rte.mc.filename[FN_MC_MOMTRNC]);
      
      if (iv+1<100)
	strcat (momcld_filename, "0");
      
      if (iv+1<10) 
	strcat (momcld_filename, "0");
      
      sprintf (tempstr, "%d", iv+1);
      
      strcat (momcld_filename, tempstr);
      strcat (momcld_filename, ".mie");
      
      break;
      
    case MOMCLD_NONE:
      break;
      
    default:
      fprintf (stderr, "Error, unknown momcld option, %d\n", input.rte.mc.momcld);
      return -1;
    }
  }

#endif

  if (verbose)
    start = clock();

  switch (input.rte.solver) {
  case MONTECARLO:
#if HAVE_MYSTIC

    /* allocate memory for 3D result fields */
    if (input.rte.mc.passback3D) {
      nres3d = input.rte.mc.Nx_sample * input.rte.mc.Ny_sample * input.atm.nzout;
      nabs3d = input.rte.mc.Nx_cloud  * input.rte.mc.Ny_cloud  * input.rte.mc.Nz_cloud;
      
      tmp_rfldir3d   = calloc (nres3d, sizeof(float));
      tmp_rfldn3d    = calloc (nres3d, sizeof(float));
      tmp_flup3d     = calloc (nres3d, sizeof(float));
      tmp_radiance3d = calloc (nres3d, sizeof(float));

      if (tmp_rfldir3d==NULL || tmp_rfldn3d==NULL || tmp_flup3d==NULL || tmp_radiance3d==NULL) {
	fprintf (stderr, "Error allocating memory for 3D result fields!\n");
	return -1;
      }

      if (input.rte.mc.absorption!=MCABS_NONE) {
	tmp_abs3d   = calloc (nabs3d, sizeof(float));
	if (tmp_abs3d==NULL) {
	  fprintf (stderr, "Error allocating memory for 3D result fields!\n");
	  return -1;
	}
      }
    }

    /* create rpv arrays */
    if (output->surf.n > 0) {
      rpv_rho0  = calloc (output->surf.n, sizeof(float));
      rpv_k     = calloc (output->surf.n, sizeof(float));
      rpv_theta = calloc (output->surf.n, sizeof(float));

      for (il=0; il<output->surf.n; il++) {
	rpv_rho0 [il] = output->surf.rpv[il].rho0_t [iv];
	rpv_k    [il] = output->surf.rpv[il].k_t    [iv];
	rpv_theta[il] = output->surf.rpv[il].theta_t[iv];
      }
    }
    else {
      rpv_rho0  = calloc (1, sizeof(float));
      rpv_k     = calloc (1, sizeof(float));
      rpv_theta = calloc (1, sizeof(float));

      rpv_rho0 [0] = output->rpv.rho0_t [iv];
      rpv_k    [0] = output->rpv.k_t    [iv];
      rpv_theta[0] = output->rpv.theta_t[iv];

    }

    switch (input.source) {
    case SRC_SOLAR: /* solar source */

      if (strlen(momcld_filename)>0)
	if (!input.quiet)
	  fprintf (stderr, " ... reading phase function from %s\n", momcld_filename);

      source = MCSRC_SOLAR;
      nmom = output->aer.optprop.nmom[iv][output->atm.nlyr-1];

      status = mystic (&output->atm.nlyr,
		       output->mc.dtair, output->mc.dtmol, 
		       output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer, 
		       output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
		       output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
		       output->mc.lwccld, output->mc.recld,
		       output->mc.temper,
		       output->mc.z,
		       &(output->surface_temperature),
		       input.rte.umu, &input.rte.numu, 
		       input.rte.phi, &input.rte.nphi, 
		       input.rte.mc.azimuth,
		       momcld_filename,
		       input.rte.mc.filename[FN_MC_PCLD],
		       output->mc.momaer, output->mc.nmomaer,
		       &output->atm.albedo_t[iv], 
		       &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
		       &(output->mc_photons),
		       &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
		       &(input.rte.mc.Nx_sample), &(input.rte.mc.Ny_sample), 
		       output->atm.zout_sur, &(output->atm.nzout),
		       rpv_rho0, rpv_k, rpv_theta,
		       &(input.rte.mc.ipa), &(input.rte.mc.escape), &(input.rte.mc.absorption),
		       &mc_loaddata,
		       input.rte.mc.filename[FN_MC_BASENAME], 
		       input.rte.mc.filename[FN_MC_WCLOUD], 
		       input.rte.mc.filename[FN_MC_ALBEDO], 
		       input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
		       input.rte.mc.filename[FN_MC_ELEVATION], 
		       &(input.rte.mc.truncate),
		       &(input.rte.mc.bcond),
		       &(input.quiet),
		       rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		       rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
		       tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
		       tmp_abs3d, 
		       &(input.rte.mc.Nx_cloud), 
		       &(input.rte.mc.Ny_cloud), 
		       &(input.rte.mc.Nz_cloud));
      
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by mystic()\n", status);
	return status;
      }
      
      if (input.rte.mc.passback3D) {

	status = cp_3d_to_rte_output (rte_out, 
				      tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
				      input.atm.nzout, 
				      input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);

	if (status!=0) {
	  fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
	  return status;
	}

	if (input.rte.mc.absorption!=MCABS_NONE) {
	  status = cp_abs_to_rte_output (rte_out, 
					 tmp_abs3d,
					 input.rte.mc.Nz_cloud,
					 input.rte.mc.Nx_cloud, input.rte.mc.Ny_cloud);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by cp_abs_to_rte_output()\n", status);
	    return status;
	  }
	}
      }

      break;

    case SRC_THERMAL: /* thermal source */

      thermal_photons = 0.5 * output->mc_photons;

      /* thermal emission of the atmosphere */
      fprintf (stderr, " ... thermal emission of the atmosphere\n");

      fprintf (stderr, " ... reading phase function from %s\n", momcld_filename);
      source = MCSRC_THERMAL_ATMOSPHERE;
      nmom = output->aer.optprop.nmom[iv][output->atm.nlyr-1];

      status = mystic (&output->atm.nlyr,
		       output->mc.dtair, output->mc.dtmol, 
		       output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer, 
		       output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
		       output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
		       output->mc.lwccld, output->mc.recld,
		       output->mc.temper,
		       output->mc.z,
		       &(output->surface_temperature),
		       input.rte.umu, &input.rte.numu, 
		       input.rte.phi, &input.rte.nphi, 
		       input.rte.mc.azimuth,
		       momcld_filename,
		       input.rte.mc.filename[FN_MC_PCLD],
		       output->mc.momaer, output->mc.nmomaer,
		       &output->atm.albedo_t[iv], 
		       &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
		       &thermal_photons,
		       &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
		       &(input.rte.mc.Nx_sample), &(input.rte.mc.Ny_sample), 
		       output->atm.zout_sur, &(output->atm.nzout),
		       rpv_rho0, rpv_k, rpv_theta,
		       &(input.rte.mc.ipa), &(input.rte.mc.escape), &(input.rte.mc.absorption),
		       &mc_loaddata,
		       input.rte.mc.filename[FN_MC_BASENAME],  
		       input.rte.mc.filename[FN_MC_WCLOUD], 
		       input.rte.mc.filename[FN_MC_ALBEDO], 
		       input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
		       input.rte.mc.filename[FN_MC_ELEVATION], 
		       &(input.rte.mc.truncate),
		       &(input.rte.mc.bcond),
		       &(input.quiet),
		       rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		       rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
		       tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
		       tmp_abs3d,
		       &(input.rte.mc.Nx_cloud), 
		       &(input.rte.mc.Ny_cloud), 
		       &(input.rte.mc.Nz_cloud));
      /* normal atmospheric + surface emission */
      switch (input.rte.mc.absorption) {   
      case MCABS_NONE:
      case MCABS_ABSORPTION:
      case MCABS_HEATING:
	
	thermal_photons = 0.5 * output->mc_photons;
	
	/* thermal emission of the atmosphere */
	fprintf (stderr, " ... thermal emission of the atmosphere\n");
	
	if (strlen(momcld_filename)>0)
	  fprintf (stderr, " ... reading phase function from %s\n", momcld_filename);
	
	source = MCSRC_THERMAL_ATMOSPHERE;
	nmom   = output->aer.optprop.nmom[iv][output->atm.nlyr-1];

	status = mystic (&output->atm.nlyr,
			 output->mc.dtair, output->mc.dtmol, 
			 output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer, 
			 output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
			 output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
			 output->mc.lwccld, output->mc.recld,
			 output->mc.temper,
			 output->mc.z,
			 &(output->surface_temperature),
			 input.rte.umu, &input.rte.numu, 
			 input.rte.phi, &input.rte.nphi, 
			 input.rte.mc.azimuth,
			 momcld_filename,
			 input.rte.mc.filename[FN_MC_PCLD],
			 output->mc.momaer, output->mc.nmomaer,
			 &output->atm.albedo_t[iv], 
			 &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
			 &thermal_photons,
			 &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
			 &(input.rte.mc.Nx_sample), &(input.rte.mc.Ny_sample), 
			 output->atm.zout_sur, &(output->atm.nzout),
			 rpv_rho0, rpv_k, rpv_theta,
			 &(input.rte.mc.ipa), &(input.rte.mc.escape), &(input.rte.mc.absorption),
			 &mc_loaddata,
			 input.rte.mc.filename[FN_MC_BASENAME], 
			 input.rte.mc.filename[FN_MC_WCLOUD], 
			 input.rte.mc.filename[FN_MC_ALBEDO], 
			 input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
			 input.rte.mc.filename[FN_MC_ELEVATION], 
			 &(input.rte.mc.truncate),
			 &(input.rte.mc.bcond),
			 &(input.quiet),
			 rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			 rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
			 tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
			 tmp_abs3d,
			 &(input.rte.mc.Nx_cloud), 
			 &(input.rte.mc.Ny_cloud), 
			 &(input.rte.mc.Nz_cloud));

      
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mc()\n", status);
	  return status;
	}
      
	if (input.rte.mc.passback3D) {
	  status = cp_3d_to_rte_output (rte_out, 
					tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
					input.atm.nzout, 
					input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);

	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
	    return status;
	  }

	  if (input.rte.mc.absorption!=MCABS_NONE) {
	    status = cp_abs_to_rte_output (rte_out, 
					   tmp_abs3d,
					   input.rte.mc.Nz_cloud,
					   input.rte.mc.Nx_cloud, input.rte.mc.Ny_cloud);
	  
	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by cp_abs_to_rte_output()\n", status);
	      return status;
	    }
	  }
	}
      
	/* thermal emission of the surface */

	fprintf (stderr, " ... thermal emission of the surface\n");

	/* data have already been loaded during the last MYSTIC call */
	mc_loaddata=0;

	tmp_out = calloc_rte_output (input);
	source = MCSRC_THERMAL_SURFACE;
	nmom   = output->aer.optprop.nmom[iv][output->atm.nlyr-1];

	if (strlen(momcld_filename)>0)
	  fprintf (stderr, " ... reading phase function from %s\n", momcld_filename);

	status = mystic (&output->atm.nlyr,
			 output->mc.dtair, output->mc.dtmol, 
			 output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer,  
			 output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
			 output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
			 output->mc.lwccld, output->mc.recld,
			 output->mc.temper,
			 output->mc.z,
			 &(output->surface_temperature),
			 input.rte.umu, &input.rte.numu, 
			 input.rte.phi, &input.rte.nphi, 
			 input.rte.mc.azimuth,
			 momcld_filename,
			 input.rte.mc.filename[FN_MC_PCLD],
			 output->mc.momaer, output->mc.nmomaer,
			 &output->atm.albedo_t[iv], 
			 &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
			 &thermal_photons,
			 &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
			 &(input.rte.mc.Nx_sample), &(input.rte.mc.Ny_sample), 
			 output->atm.zout_sur, &(output->atm.nzout),
			 rpv_rho0, rpv_k, rpv_theta,
			 &(input.rte.mc.ipa), &(input.rte.mc.escape), &(input.rte.mc.absorption),
			 &mc_loaddata,
			 input.rte.mc.filename[FN_MC_BASENAME], 
			 input.rte.mc.filename[FN_MC_WCLOUD], 
			 input.rte.mc.filename[FN_MC_ALBEDO], 
			 input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
			 input.rte.mc.filename[FN_MC_ELEVATION], 
			 &(input.rte.mc.truncate),
			 &(input.rte.mc.bcond),
			 &(input.quiet),
			 tmp_out->rfldir, tmp_out->rfldn, tmp_out->flup, 
			 tmp_out->uavgso, tmp_out->uavgdn, tmp_out->uavgup,
			 tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
			 tmp_abs3d,
			 &(input.rte.mc.Nx_cloud), 
			 &(input.rte.mc.Ny_cloud), 
			 &(input.rte.mc.Nz_cloud));


      
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mc()\n", status);
	  return status;
	}
      
	if (input.rte.mc.passback3D) {
	  status = cp_3d_to_rte_output (tmp_out, 
					tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
					input.atm.nzout, 
					input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);

	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
	    return status;
	  }

	  if (input.rte.mc.absorption!=MCABS_NONE) {
	    status = cp_abs_to_rte_output (tmp_out, 
					   tmp_abs3d,
					   input.rte.mc.Nz_cloud,
					   input.rte.mc.Nx_cloud, input.rte.mc.Ny_cloud);
	  
	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by cp_abs_to_rte_output()\n", status);
	      return status;
	    }
	  }
	}

	/* add atmosphere and surface contributions to get total (rte_out) */
	status = add_rte_output (rte_out, tmp_out, 1.0, input);
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by add_rte_output()\n", status);
	  return status;
	}
      
	free_rte_output (tmp_out, input);

	break;

      case MCABS_EMISSION:
	
	thermal_photons = 0;
	
	/* 3D emission field */
	fprintf (stderr, " ... 3D emission field\n");
	
	source = MCSRC_THERMAL_ATMOSPHERE;
	nmom   = output->aer.optprop.nmom[iv][output->atm.nlyr-1];
	
	status = mystic (&output->atm.nlyr,
			 output->mc.dtair, output->mc.dtmol, 
			 output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer, 
			 output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
			 output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
			 output->mc.lwccld, output->mc.recld,
			 output->mc.temper,
			 output->mc.z,
			 &(output->surface_temperature),
			 input.rte.umu, &input.rte.numu, 
			 input.rte.phi, &input.rte.nphi, 
			 input.rte.mc.azimuth,
			 momcld_filename,
			 input.rte.mc.filename[FN_MC_PCLD],
			 output->mc.momaer, output->mc.nmomaer,
			 &output->atm.albedo_t[iv], 
			 &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
			 &thermal_photons,
			 &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
			 &(input.rte.mc.Nx_sample), &(input.rte.mc.Ny_sample), 
			 output->atm.zout_sur, &(output->atm.nzout),
			 rpv_rho0, rpv_k, rpv_theta,
			 &(input.rte.mc.ipa), &(input.rte.mc.escape), &(input.rte.mc.absorption),
			 &mc_loaddata,
			 input.rte.mc.filename[FN_MC_BASENAME], 
			 input.rte.mc.filename[FN_MC_WCLOUD], 
			 input.rte.mc.filename[FN_MC_ALBEDO], 
			 input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
			 input.rte.mc.filename[FN_MC_ELEVATION], 
			 &(input.rte.mc.truncate),
			 &(input.rte.mc.bcond),
			 &(input.quiet),
			 rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			 rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
			 tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
			 tmp_abs3d,
			 &(input.rte.mc.Nx_cloud), 
			 &(input.rte.mc.Ny_cloud), 
			 &(input.rte.mc.Nz_cloud));

      
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mc()\n", status);
	  return status;
	}
	
	if (input.rte.mc.passback3D) {
	  status = cp_3d_to_rte_output (rte_out, 
					tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, tmp_radiance3d,
					input.atm.nzout, 
					input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);

	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
	    return status;
	  }

	  if (input.rte.mc.absorption!=MCABS_NONE) {
	    status = cp_abs_to_rte_output (rte_out, 
					   tmp_abs3d,
					   input.rte.mc.Nz_cloud,
					   input.rte.mc.Nx_cloud, input.rte.mc.Ny_cloud);
	  
	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by cp_abs_to_rte_output()\n", status);
	      return status;
	    }
	  }
	}

	break;

      default:
	fprintf (stderr, "Error, unknown absorption type %d\n", input.rte.mc.absorption);
	return -1;
      }

      break;

    default:
      fprintf (stderr, "Error, unknown source\n");
      return -1;
    }


    /* free memory of 3D results */
    if (input.rte.mc.passback3D) {
      free (tmp_rfldir3d);
      free (tmp_rfldn3d);
      free (tmp_flup3d);

      if (tmp_radiance3d != NULL)
	free (tmp_radiance3d);

      if (input.rte.mc.absorption!=MCABS_NONE)
	free (tmp_abs3d);
    }

    /* free RPV arrays */
    free (rpv_rho0);
    free (rpv_k);
    free (rpv_theta);


    break;

#else
    fprintf (stderr, "Error: MYSTIC solver not included in uvspec build.\n");
    fprintf (stderr, "Error: Please contact bernhard.mayer@dlr.de\n");
    return -1;
#endif	

  case DISORT:
  case DISORT2:

    disort_u0u  = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu   = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    if ( input.rte.solver == DISORT ) {

      status = F77_FCN (dcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
				&input.rte.numu, &input.rte.nphi); 
	
      if (status!=0)
	return status;


      disort_pmom = c2fortran_2D_float_ary(output->atm.nlyr, input.rte.nstr+1,   output->pmom);
      F77_FCN  (disort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
			output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
			&(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &input.atm.nzout, rte_in->utau,
			&input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
			&input.rte.nphi, input.rte.phi, &(rte_in->ibcnd), &(rte_in->fbeam), 
			&(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->fisot), &(rte_in->lamber), 
			&output->atm.albedo_t[iv], rte_in->hl, &(rte_in->btemp), &(rte_in->ttemp), 
			&(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
			&(rte_in->onlyfl), &(rte_in->accur), 
			rte_in->prndis, rte_in->header, &output->atm.nlyr, 
			&input.atm.nzout, &input.rte.maxumu, &input.rte.nstr,
			&input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			rte_out->dfdt, rte_out->uavg, 
			disort_uu, disort_u0u, rte_out->albmed, rte_out->trnmed, 
			rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup );
    }
    else if ( input.rte.solver == DISORT2 ) {
      
      status = F77_FCN (dcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
				 &input.rte.numu, &input.rte.nphi); 
      
      if (status!=0)
	return status;
      

      /* BRDF or Lambertian albedo */
      if (input.brdf != BRDF_NONE)
	rte_in->lamber = 0;
      else
	rte_in->lamber = 1;
	
      disort2_pmom = c2fortran_2D_float_ary(output->atm.nlyr, output->atm.nmom+1, output->pmom);

      F77_FCN  (disort2)(&output->atm.nlyr, output->dtauc, output->ssalb, &output->atm.nmom,
			 disort2_pmom, output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
			 &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &input.atm.nzout, rte_in->utau,
			 &input.rte.nstr, &rte_in->usrang, &input.rte.numu, input.rte.umu, 
			 &input.rte.nphi, input.rte.phi, &rte_in->ibcnd, &rte_in->fbeam, 
			 &rte_in->umu0, &output->atm.phi0_t[iv], &rte_in->fisot, &rte_in->lamber, 
			 &output->atm.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
			 &rte_in->temis, &rte_in->planck,
			 &rte_in->onlyfl, &rte_in->accur,
			 rte_in->prndis2, rte_in->header, &output->atm.nlyr, 
			 &input.atm.nzout, &input.rte.maxumu, &input.rte.maxphi, &output->atm.nmom,
			 rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			 rte_out->dfdt, rte_out->uavg, 
			 disort_uu, rte_out->albmed, rte_out->trnmed, 
			 rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
			 &(input.brdf),
			 &(output->rpv.rho0_t[iv]), &(output->rpv.k_t[iv]), &(output->rpv.theta_t[iv]),
			 &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal));
    }    
    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }

    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (input.atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, input.atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(disort_pmom);
    free(disort2_pmom);
    free(disort_u0u);
    free(disort_uu);

    break;

  case TZS:

    tzs_u0u  = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
    tzs_uu   = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    status = F77_FCN (dcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0)
      return status;
    
    
    /* call RTE solver */
    F77_FCN  (tzs) (&output->atm.nlyr, output->dtauc, output->ssalb, 
		    output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
		    &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &input.atm.nzout, rte_in->utau,
		    &rte_in->usrang, &input.rte.numu, input.rte.umu, 
		    &input.rte.nphi, input.rte.phi, 
		    &output->atm.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
		    &rte_in->temis, &rte_in->planck,
		    rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		    &input.atm.nzout, &input.rte.maxumu, &input.rte.maxphi, 
		    rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		    rte_out->dfdt, rte_out->uavg, 
		    tzs_uu, rte_out->albmed, rte_out->trnmed, 
		    rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup);

    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }

    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (input.atm.nzout, 
				    input.rte.maxumu, tzs_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, input.atm.nzout, 
				    input.rte.maxumu, tzs_uu, rte_out->uu);

    free(tzs_u0u);
    free(tzs_uu);

    break;

  case SSS:

    sss_u0u  = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
    sss_uu   = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    status = F77_FCN (dcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by dcheck()\n", status);
      return status;
    }
    
    /* call RTE solver */
    sss_pmom = c2fortran_2D_float_ary(output->atm.nlyr, output->atm.nmom+1, output->pmom);
    F77_FCN  (sss)(&output->atm.nlyr, output->dtauc, output->ssalb, &output->atm.nmom,
		   sss_pmom, output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
		   &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &input.atm.nzout, rte_in->utau,
		   &input.rte.nstr, &rte_in->usrang, &input.rte.numu, input.rte.umu, 
		   &input.rte.nphi, input.rte.phi, &rte_in->ibcnd, &rte_in->fbeam, 
		   &rte_in->umu0, &output->atm.phi0_t[iv], &rte_in->fisot, &rte_in->lamber, 
		   &output->atm.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
		   &rte_in->temis, &rte_in->planck,
		   &rte_in->onlyfl, &rte_in->accur,
		   rte_in->prndis2, rte_in->header, &output->atm.nlyr, 
		   &input.atm.nzout, &input.rte.maxumu, &input.rte.maxphi, &output->atm.nmom,
		   rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		   rte_out->dfdt, rte_out->uavg, 
		   sss_uu, rte_out->albmed, rte_out->trnmed, 
		   rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		   &(input.brdf),
		   &(output->rpv.rho0_t[iv]), &(output->rpv.k_t[iv]), &(output->rpv.theta_t[iv]),
		   &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal));
    
    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }


    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (input.atm.nzout, 
				    input.rte.maxumu, sss_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, input.atm.nzout, 
				    input.rte.maxumu, sss_uu, rte_out->uu);

    free(sss_pmom);
    free(sss_u0u);
    free(sss_uu);

    break;

  case SSSI:

#if HAVE_SSSI

    /* get cloud reflectivity from lookup table */
    status = read_isccp_reflectivity (output->sssi.type, output->atm.sza_t[iv], output->sssi.tautot, 
				      input.filename[FN_PATH], input.quiet, &(output->sssi.ref));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_isccp_reflectivity()\n", status);
      return status;
    }

    if (input.verbose) {
      fprintf (stderr, "*** SSSI cloud properties:\n");
      fprintf (stderr, "    top level:          zd[%d] = %f\n", 
	       output->sssi.lctop, output->atm.zd[output->sssi.lctop]);
      switch (output->sssi.type) {
      case ISCCP_WATER: 
	fprintf (stderr, "    type:               water\n");
	break;
      case ISCCP_ICE: 
	fprintf (stderr, "    type:               ice\n");
	break;
      default:
	fprintf (stderr, "Error, unknown ISCCP cloud type %d\n", output->sssi.type);
	return -1;
      }
      fprintf (stderr, "    optical thickness:  %f\n", 
	       output->sssi.tautot);
      fprintf (stderr, "    reflectivity:       %f\n", 
	       output->sssi.ref);
    }
    fflush (stderr);

    sss_u0u  = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
    sss_uu   = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    status = F77_FCN (dcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by dcheck()\n", status);
      return status;
    }
    
    /* call RTE solver */
    sss_pmom = c2fortran_2D_float_ary(output->atm.nlyr, output->atm.nmom+1, output->pmom);
    F77_FCN  (sssi)(&output->atm.nlyr, output->dtauc, output->ssalb, &output->atm.nmom,
		    sss_pmom, output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
		    &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &input.atm.nzout, rte_in->utau,
		    &input.rte.nstr, &rte_in->usrang, &input.rte.numu, input.rte.umu, 
		    &input.rte.nphi, input.rte.phi, &rte_in->ibcnd, &rte_in->fbeam, 
		    &rte_in->umu0, &output->atm.phi0_t[iv], &rte_in->fisot, &rte_in->lamber, 
		    &output->atm.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
		    &rte_in->temis, &rte_in->planck,
		    &rte_in->onlyfl, &rte_in->accur,
		    rte_in->prndis2, rte_in->header, &output->atm.nlyr, 
		    &input.atm.nzout, &input.rte.maxumu, &input.rte.maxphi, &output->atm.nmom,
		    rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		    rte_out->dfdt, rte_out->uavg, 
		    sss_uu, rte_out->albmed, rte_out->trnmed, 
		    rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		    &(input.brdf),
		    &(output->rpv.rho0_t[iv]), &(output->rpv.k_t[iv]), &(output->rpv.theta_t[iv]),
		    &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), 
		    &(output->sssi.ref), &(output->sssi.lctop));
    
    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }


    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (input.atm.nzout, 
				    input.rte.maxumu, sss_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, input.atm.nzout, 
				    input.rte.maxumu, sss_uu, rte_out->uu);

    free(sss_pmom);
    free(sss_u0u);
    free(sss_uu);
#else
  fprintf (stderr, "Error, SSSI solver not available!\n");
  return -1;
#endif

    break;

  case POLRADTRAN:
#if HAVE_POLRADTRAN
    rte_in->pol.nummu    = input.rte.nstr/2+input.rte.numu;
    rte_in->pol.albedo   = (double) output->atm.albedo_t[iv];
    rte_in->pol.btemp    = (double) (rte_in->btemp);
    rte_in->pol.flux     = (double) (rte_in->fbeam*rte_in->umu0); /* Polradtran wants flux 
                                                                     on horizontal surface */
    rte_in->pol.mu       = (double) (rte_in->umu0);
    rte_in->pol.sky_temp = (double) (rte_in->ttemp);
    rte_in->pol.wavelength = 0.0; /* Not used for solar source */
    for (lu=0; lu<output->atm.nlyr; lu++) {
      /* rte_in->pol.gas_extinct[lu] = (double) (1.-output->ssalb[lu])*output->dtauc[lu]; */
      rte_in->pol.gas_extinct[lu] = 0.0;
    }

    polradtran_down_flux = (double *) calloc (input.atm.nzout*input.rte.pol_nstokes, sizeof(double));
    polradtran_up_flux   = (double *) calloc (input.atm.nzout*input.rte.pol_nstokes, sizeof(double));
    polradtran_down_rad  = (double *) calloc (input.atm.nzout*(input.rte.pol_aziorder+1)*
                         		      (input.rte.nstr/2+input.rte.numu)*input.rte.pol_nstokes,
					      sizeof(double));
    polradtran_up_rad    = (double *) calloc (input.atm.nzout*(input.rte.pol_aziorder+1)*
					      (input.rte.nstr/2+input.rte.numu)*input.rte.pol_nstokes,
					      sizeof(double));
    for (iu=0; iu<rte_in->pol.nummu; iu++) 
      rte_out->polradtran_mu_values[iu] = (double) fabs(output->mu_values[iu]);
    /* Take fabs since radtran will calculate both up_rad and down_rad for the value of mu.  */
    /* The user wants one of these. That is sorted out in the output section of uvspec_lex.l */

    F77_FCN (radtran) (&input.rte.pol_nstokes, &(rte_in->pol.nummu), 
		       &input.rte.pol_aziorder, &input.rte.pol_max_delta_tau,
		       &input.rte.pol_src_code, input.rte.pol_quad_type,
		       rte_in->pol.deltam, &(rte_in->pol.flux), &(rte_in->pol.mu),
		       &(rte_in->pol.btemp), (rte_in->pol.ground_type),
		       &(rte_in->pol.albedo), &(rte_in->pol.ground_index),
		       &(rte_in->pol.sky_temp), &(rte_in->pol.wavelength),
		       &output->atm.nlyr, (rte_in->pol.height), 
		       (rte_in->pol.temperatures), (rte_in->pol.gas_extinct),
		       output->atm.pol_scat_files, &input.atm.nzout,
		       (rte_in->pol.outlevels), rte_out->polradtran_mu_values,
		       polradtran_up_flux, polradtran_down_flux,
		       polradtran_up_rad, polradtran_down_rad);
    
    fortran2c_2D_double_ary_noalloc (input.atm.nzout, 
				     input.rte.pol_nstokes, 
				     polradtran_up_flux,
				     rte_out->polradtran_up_flux);
    
    fortran2c_2D_double_ary_noalloc (input.atm.nzout, 
				     input.rte.pol_nstokes, 
				     polradtran_down_flux,
				     rte_out->polradtran_down_flux);

    if (input.rte.pol_aziorder > 0 ) {
      fortran2c_4D_double_ary_noalloc (input.atm.nzout, 
				       input.rte.pol_aziorder+1,
				       input.rte.nstr/2+input.rte.numu, 
				       input.rte.pol_nstokes, 
				       polradtran_down_rad,
				       rte_out->polradtran_down_rad_q);
      
      fortran2c_4D_double_ary_noalloc (input.atm.nzout, 
				       input.rte.pol_aziorder+1,
				       input.rte.nstr/2+input.rte.numu, 
				       input.rte.pol_nstokes, 
				       polradtran_up_rad,
				       rte_out->polradtran_up_rad_q);

      fourier2azimuth (rte_out->polradtran_down_rad_q, rte_out->polradtran_up_rad_q,
		       rte_out->polradtran_down_rad, rte_out->polradtran_up_rad, 
		       input.atm.nzout, input.rte.pol_aziorder,
		       input.rte.nstr, input.rte.numu, input.rte.pol_nstokes,
		       input.rte.nphi, input.rte.phi);
    }


    free(polradtran_down_flux);
    free(polradtran_up_flux);

    free(polradtran_down_rad);
    free(polradtran_up_rad);

#else
    fprintf (stderr, "Error: RTE polradtran solver not included in uvspec build.\n");
    fprintf (stderr, "Error: Get solver and rebuild uvspec.\n");
    return -1;
#endif	
    break;
  case QDISORT:

    status = F77_FCN (dcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
			       &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0)
      return status;

    sdisort_beta = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    disort_pmom  = c2fortran_2D_float_ary(output->atm.nlyr, input.rte.nstr+1, output->pmom );
    disort_u0u   = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu    = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));
    sdisort_sig  = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    if ((status  = ASCII_calloc_float(&sdisort_denstab, output->atm.microphys.nsza_denstab, output->atm.nlyr+1)) !=0)
      return status;
    if (output->atm.microphys.denstab_id > 0) {
      for (lu=0; lu<=output->atm.nlyr; lu++) {
	sdisort_sig[lu] = output->crs.crs_amf[iv][lu];
	for (is=0; is<output->atm.microphys.nsza_denstab; is++)
	  sdisort_denstab[is][lu] = output->atm.microphys.denstab_amf[is][lu];
      }

      tosdisort_denstab = c2fortran_2D_float_ary (output->atm.microphys.nsza_denstab, 
						  output->atm.nlyr+1, sdisort_denstab);
    }

    F77_FCN (qdisort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
		      output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
		      &(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &input.atm.nzout, rte_in->utau,
		      &input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
		      &input.rte.nphi, input.rte.phi, &(rte_in->fbeam), sdisort_beta, &(rte_in->nil),
		      &(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->newgeo), output->atm.zd,
		      &(rte_in->spher), &radius, &(rte_in->fisot), &output->atm.albedo_t[iv], 
		      &(rte_in->btemp), &(rte_in->ttemp), &(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
		      &(rte_in->onlyfl), &(rte_in->accur), &(rte_in->quiet), rte_in->ierror_s, 
		      rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		      &input.atm.nzout, &input.rte.maxumu, &input.rte.nstr,
		      &input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		      rte_out->dfdt, rte_out->uavg,
		      disort_uu, disort_u0u, &input.rte.nscat,
		      rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		      &input.rte.nrefrac, &input.rte.ichap, output->atm.microphys.refind,
		      &output->atm.microphys.nsza_denstab, output->atm.microphys.sza_denstab,
 		      sdisort_sig, tosdisort_denstab, output->dtauc_md );

    /* convert temporary Fortran arrays to permanent result */

    fortran2c_2D_float_ary_noalloc (input.atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, input.atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(sdisort_beta);
    free(disort_pmom);
    free(disort_u0u);
    free(disort_uu);

    break;
  case SDISORT:

    status = F77_FCN (dcheck) (&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
			       &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0)
      return status;

    sdisort_beta = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    disort_pmom  = c2fortran_2D_float_ary(output->atm.nlyr, input.rte.nstr+1, output->pmom );
    disort_u0u   = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu    = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));
    sdisort_sig  = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    if ((status  = ASCII_calloc_float(&sdisort_denstab, output->atm.microphys.nsza_denstab, output->atm.nlyr+1)) !=0)
      return status;
    if (output->atm.microphys.denstab_id > 0) {
      for (lu=0; lu<=output->atm.nlyr; lu++) {
	sdisort_sig[lu] = output->crs.crs_amf[iv][lu];
	for (is=0; is<output->atm.microphys.nsza_denstab; is++)
	  sdisort_denstab[is][lu] = output->atm.microphys.denstab_amf[is][lu];
      }

      tosdisort_denstab = c2fortran_2D_float_ary (output->atm.microphys.nsza_denstab, 
						  output->atm.nlyr+1, sdisort_denstab);
    }

    F77_FCN (sdisort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
		      output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
		      &(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &input.atm.nzout, rte_in->utau,
		      &input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
		      &input.rte.nphi, input.rte.phi, &(rte_in->fbeam), sdisort_beta, &(rte_in->nil),
		      &(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->newgeo), output->atm.zd,
		      &(rte_in->spher), &radius, &(rte_in->fisot), &output->atm.albedo_t[iv], 
		      &(rte_in->btemp), &(rte_in->ttemp), &(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
		      &(rte_in->onlyfl), &(rte_in->accur), &(rte_in->quiet), rte_in->ierror_s, 
		      rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		      &input.atm.nzout, &input.rte.maxumu, &input.rte.nstr,
		      &input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		      rte_out->dfdt, rte_out->uavg,
		      disort_uu, disort_u0u, &input.rte.nscat,
		      rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		      &input.rte.nrefrac, &input.rte.ichap, output->atm.microphys.refind,
		      &output->atm.microphys.nsza_denstab, output->atm.microphys.sza_denstab,
 		      sdisort_sig, tosdisort_denstab, output->dtauc_md );

    /* convert temporary Fortran arrays to permanent result */

    fortran2c_2D_float_ary_noalloc (input.atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, input.atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(sdisort_beta);
    free(disort_pmom);
    free(disort_u0u);
    free(disort_uu);

    break;
  case SPSDISORT:

    status = F77_FCN (dcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
			       &input.rte.numu, &input.rte.nphi); 
	
    if (status!=0)
      return status;

    sdisort_beta = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    disort_pmom = c2fortran_2D_float_ary(output->atm.nlyr, input.rte.nstr+1, output->pmom );
    disort_u0u  = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu   = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    F77_FCN (spsdisort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
		      output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
		      &(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &input.atm.nzout, rte_in->utau,
		      &input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
		      &input.rte.nphi, input.rte.phi, &(rte_in->fbeam), sdisort_beta, &(rte_in->nil),
		      &(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->newgeo), output->atm.zd,
		      &(rte_in->spher), &radius, &(rte_in->fisot), &output->atm.albedo_t[iv], 
		      &(rte_in->btemp), &(rte_in->ttemp), &(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
		      &(rte_in->onlyfl), &(rte_in->accur), &(rte_in->quiet), rte_in->ierror_s, 
		      rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		      &input.atm.nzout, &input.rte.maxumu, &input.rte.nstr,
		      &input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		      rte_out->dfdt, rte_out->uavg, 
		      disort_uu, disort_u0u, 
		      rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup );

    /* convert temporary Fortran arrays to permanent result */

    fortran2c_2D_float_ary_noalloc (input.atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, input.atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(sdisort_beta);
    free(disort_pmom);
    free(disort_u0u);
    free(disort_uu);

    break;
  case TWOSTR:

    status = F77_FCN (tcheck)(&output->atm.nlyr, &input.atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0)
      return status;

    twostr_gg = (float *) calloc (output->atm.nlyr, sizeof(float));
    for (lu=0; lu<output->atm.nlyr; lu++) 
      twostr_gg[lu]=output->pmom[lu][1];
		

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    F77_FCN  (twostr)( &output->atm.albedo_t[iv], &(rte_in->btemp), &input.rte.deltam,
		       output->dtauc, &(rte_in->fbeam), &(rte_in->fisot), twostr_gg, rte_in->header, 
		       rte_in->ierror_t, &output->atm.nlyr, &input.atm.nzout, &(rte_in->newgeo), 
		       &output->atm.nlyr, &(rte_in->planck), &input.atm.nzout,
		       rte_in->prntwo, &(rte_in->quiet), &radius, &(rte_in->spher), 
		       output->ssalb, &(rte_in->temis),  
		       output->atm.microphys.temper, &(rte_in->ttemp), &(rte_in->umu0), 
		       &(rte_in->usrtau), rte_in->utau, &(output->wl.wvnmlo_t[iv]),
		       &(output->wl.wvnmhi_t[iv]), output->atm.zd, rte_out->dfdt, rte_out->flup, 
		       rte_out->rfldir, rte_out->rfldn, rte_out->uavg);
    free(twostr_gg);

    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }

    break;
  case SUCORD:
#if HAVE_SOS
    status = sos (output->atm.nlyr, rte_in->newgeo, input.rte.nstr, 
		  input.rte.nscat, output->atm.albedo_t[iv],
		  radius, output->atm.zd, output->ssalb, output->pmom,
		  output->dtauc, 
		  output->atm.sza_t[iv], input.atm.nzout, 
		  input.rte.numu, input.rte.umu,
		  rte_in->utau, 
		  rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		  rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
		  rte_out->u0u);
		
    if (status!=0) {
      fprintf (stderr, "Error %d returned by sos()\n", status);
      return status;
    }
    break;
#else
    fprintf (stderr, "Error: sos solver not included in uvspec build.\n");
    fprintf (stderr, "Error: Please contact arve.kylling@nilu.no\n");
    return -1;
#endif	

  case NULLSOLVER:
    /* do nothing */
    break;

  default:
    fprintf (stderr, "Error: RTE solver not yet implemented\n");
    return -1;
  }

  if (verbose) {
    end = clock();
    fprintf (stderr, "*** last solver call: %f seconds\n", 
	     ((double) (end - start)) / CLOCKS_PER_SEC);
  }
      
  return 0;    /* if o.k. */
}


static int init_rte_input (rte_input *rte, input_struct input, output_struct *output)
{
  int i=0, found=0;
  int ip=0, is=0, lu=0;
  int status = 0;

  strcpy(rte->header, "");
  rte->accur = 1.e-3;


  switch (input.source) {
  case SRC_NONE:
    rte->planck = 0; 
    rte->fbeam  = 0.0;
    rte->fisot  = 0.0;
    break;

  case SRC_SOLAR:
    rte->planck=0; 

    if (input.rte.fisot > 0) { 
      rte->fbeam = 0.0; 
      rte->fisot = 1.0; 
    }
    else { 
      rte->fbeam = 1.0; 
      rte->fisot = 0.0; 
    }

    break;

  case SRC_THERMAL:
    rte->planck = 1; 
    rte->fbeam  = 0.0;
    rte->fisot  = 0.0;
    
    rte->btemp = output->surface_temperature;
    rte->ttemp = output->atm.microphys.temper[0];

    rte->temis = 0.0;
    
    break;
  default:
    fprintf (stderr, "Error, unknown source\n");
    return -1;
  }

  rte->umu0  = 0.0;

  rte->ierror_d[0] = 0;
  rte->ierror_s[0] = 0;
  rte->ierror_t[0] = 0;

  for (i=0; i<7; i++)
    rte->prndis[i]  = 0;

  for (i=0; i<5; i++)
    rte->prndis2[i] = 0;

  for (i=0; i<2; i++)
    rte->prntwo[i]  = 0;

  rte->ibcnd=0;
  rte->lamber=1;
  rte->newgeo=1;
  rte->nil=0;
  rte->onlyfl=1;
  rte->quiet = 0;
  rte->spher = 1;
  rte->usrang = 0;
  rte->usrtau = 1;


  rte->hl   = (float *) calloc (input.rte.maxumu+1, sizeof(float));
  rte->utau = (float *) calloc (input.atm.nzout,    sizeof(float));


  if (input.rte.numu > 0 ) { 
    rte->onlyfl = 0; 
    rte->usrang = 1; 
  }


  /* PolRadtran */
  rte->pol.deltam       = "Y";
  rte->pol.ground_type  = "L";
  strcpy(rte->pol.polscat, "");
  rte->pol.albedo       = 0.0;
  rte->pol.btemp        = 0.0;
  rte->pol.flux         = 1.0;
  rte->pol.gas_extinct  = NULL;
  rte->pol.height       = NULL;
  rte->pol.mu           = 1.0;
  rte->pol.sky_temp     = 0.0;
  rte->pol.temperatures = NULL;
  rte->pol.wavelength   = 0.0;
  rte->pol.outlevels    = NULL;
  rte->pol.nummu        = 0; 
  
  rte->pol.ground_index.re = 0.0;
  rte->pol.ground_index.im = 0.0;


  switch (input.rte.solver) {
  case MONTECARLO:
    /* set number of photons */
    output->mc_photons = input.rte.mc.photons;
    break;
  case SDISORT:
  case SPSDISORT:
  case QDISORT:
  case DISORT:
  case SSS:
  case SSSI:
  case TZS:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      (rte->prndis)[input.rte.prndis[ip]-1] = 1;
    break;
  case DISORT2:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      (rte->prndis2)[input.rte.prndis[ip]-1] = 1;
    break;
  case POLRADTRAN:
    /* Need to do a little checking of polradtran specific input stuff here...*/
    rte->pol.outlevels    = (int *) calloc (input.atm.nzout, sizeof(int));
    found = 0;
    for (i=0; i<input.atm.nzout; i++) 
      for (lu=0; lu<=output->atm.nlyr; lu++) 
	if (fabs(input.atm.zout_sur[i] - output->atm.zd[lu]) <= 0) 
	  rte->pol.outlevels[found++] = lu+1;

    if ( found != input.atm.nzout ) {
      fprintf (stderr, "*** zout does not correspond to atmosphere file levels.\n");
      fprintf (stderr, "*** zout must do so for the polradtran solver.\n");
      status--;
    }

    if (input.rte.deltam == 0) 
      rte->pol.deltam = "N";
    else {
      if (input.rte.deltam == 1) 
	rte->pol.deltam = "Y";
    }

    if ((output->mu_values = (float *) calloc (input.rte.nstr/2+input.rte.numu, sizeof(float))) == NULL)
      return -1;

    if ( strncmp(input.rte.pol_quad_type,"E",1)==0 )
      for (i=0; i<input.rte.numu; i++) 
	output->mu_values[input.rte.nstr/2+i] = input.rte.umu[i];


    rte->pol.gas_extinct  = (double *) calloc (output->atm.nlyr+1, sizeof(double));
    rte->pol.height       = (double *) calloc (output->atm.nlyr+1, sizeof(double));
    rte->pol.temperatures = (double *) calloc (output->atm.nlyr+1, sizeof(double));

    output->atm.pol_scat_files= (char *) calloc (64*output->atm.nlyr, sizeof(char));      

    for (is=0;is<64*output->atm.nlyr;is++)
      output->atm.pol_scat_files[is] = ' ';

    for (lu=0;lu<output->atm.nlyr;lu++) {
      is = lu*64;
      sprintf(rte->pol.polscat,".scat_file_%03d",lu);
      strcpy(&output->atm.pol_scat_files[is], rte->pol.polscat);
    }

    for (lu=0;lu<=output->atm.nlyr;lu++) {
      rte->pol.height[lu] = (double) lu;  /*  Weird hey???? Well, the story goes as 
					      follows: polradtran wants extinction and
					      scattering in terms of 1/(layerthickness).
					      However, we feed it optical depth. Hence,
					      we need to make delta_height of each layer
					      equal one. One simple remedy is the one
					      used. Arve 15.03.2000 */
      rte->pol.temperatures[lu] = (double) output->atm.microphys.temper[lu];
    }
    break;
  case TWOSTR:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      if ( input.rte.prndis[ip]==1 || input.rte.prndis[ip]==2 ) 
	(rte->prntwo)[input.rte.prndis[ip]-1] = 1;
    break;
  case SUCORD:
  case NULLSOLVER:
    break;
  default:
    fprintf (stderr, "Error: RTE solver not yet implemented\n");
    break;
  }    
  
  return status;
}

static void fourier2azimuth (double**** down_rad_rt3, double**** up_rad_rt3,
			     double**** down_rad, double**** up_rad, 
			     int nzout, int aziorder, int nstr, int numu, int nstokes,
			     int nphi, float* phi)
{
  /* For each azimuth and polar angle sum the Fourier azimuth series appropriate
     for the particular Stokes parameter to produce the radiance.
     Only used for the polradtran solver 
  */
  int i=0, j=0, je=0, k=0, lu=0, m=0;
  double sumd=0, sumu=0;
  float phir;
  for (lu=0;lu<nzout;lu++) {
    for (k=0;k<nphi;k++) {
      phir = PI*phi[k]/180.0;
      
      /* Up- and downwelling irradiances at user angles only*/
      /*      for (j=0;j<nstr/2+numu;j++) {  */
      for (j=0;j<numu;j++) {
	je = j+nstr/2;
	for (i=0;i<nstokes;i++) {
	  sumd=0.0;
	  sumu=0.0;
	  for (m=0;m<=aziorder;m++) {
	    if (i < 2 ) {
	      sumd += cos(m*phir)*down_rad_rt3[lu][m][je][i];
	      sumu += cos(m*phir)*up_rad_rt3[lu][m][je][i];
	    }
	    else {
	      sumd += sin(m*phir)*down_rad_rt3[lu][m][je][i];
	      sumu += sin(m*phir)*up_rad_rt3[lu][m][je][i];
	    }
	  }
	  down_rad[lu][k][j][i] = sumd;
	  up_rad[lu][k][j][i] = sumu;
	}
      }
    }
  }
}

//static int calc_spectral_heating (rte_output *rte_band, rte_output *rte_out,
//			      float* zout, float* dens_air, float* utau, input_struct input )
static int calc_spectral_heating (float* heat, float* dfdt,
			      float* zout, float* dens_air, float* utau, input_struct input )
{
  float dtdz = 0, rho_m=0;
  int lu=0;
  int status = 0;
  float factor  = 0;
  double M_AIR = 0.028977;  // molecular weight of air (kg mol-1)
  double N_A   = 6.022e+23; // Avogadros constant (mol-1)
  double cp = 1004;         // Specific heat at constant pressure (J kg-1 K-1)
  double s2day = 3600*24;   // seconds to day
  double mW2W  = 1.e-3;     // mW to W

  // Figure out if we have to convert from mW to W. This
  // should cover the input stuff that comes with uvspec. Now,
  // if the user decides to change the input solar flux it may get interesting..
  switch(input.ck_scheme) {
  case CK_FU:
  case CK_KATO:
  case CK_KATO2:
  case CK_KATO2_96:
  case CK_AVHRR_KRATZ:
    factor = 1.0;
    break;
  case CK_SBDART:
  case CK_GENERIC:
  case CK_NONE:
    factor = mW2W;
    break;
    
  default:
    fprintf (stderr, "Error: unsupported correlated-k scheme %d\n", input.ck_scheme);
    return -1;
    break;
  }
  
  for(lu=0; lu<input.atm.nzout; lu++) {

    /*     Compute the derivative dt/dz using a two-point formula */

    if ( lu == 0 ) {
      dtdz = (utau[lu]-utau[lu+1])/(zout[lu]-zout[lu+1]);
    }
    else {
      dtdz = (utau[lu-1]-utau[lu])/(zout[lu-1]-zout[lu]);
    }
    dtdz = dtdz * 1.e-03;   //      !  Convert from 1/km to 1/m 
    rho_m = dens_air[lu]*1.e+6*M_AIR/N_A;
   
    heat   [lu] += dfdt[lu] * (-dtdz) * factor * (s2day/(rho_m*cp));


  }

  return status;

}
