/*--------------------------------------------------------------------
 * $Id: uvspec.c,v 1.48 2004/07/23 12:12:41 arve Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include "uvspec.h"
#include "../src_c/numeric.h"


/* prototypes of internal functions */
void pmesg (char *message, int verbose);



/**************************************************************/
/* The main uvspec function.                                  */
/**************************************************************/

int uvspec (input_struct input, output_struct *output)
{
  int status = 0;

  /**** Check the model input data ****/
  pmesg (" ... calling uvspec_check(), checking model input data\n", input.verbose);
  status = uvspec_check(input);
  if (status!=0) {
    fprintf (stderr, "%d error(s) in uvspec input-file, aborting\n", abs(status));
    return status;
  }


  /**** Set wavelength grid for the transmittance calculation ****/
  pmesg (" ... calling setup_wlgrid(), generating model wavelength grid\n", input.verbose);
  status = setup_wlgrid (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up wavelength grid\n", status);
    return status;
  } 


  /**** Setup sza and phi0 ****/
  pmesg (" ... calling setup_sza(), generating model solar zenith and azimuth\n", input.verbose);
  status = setup_sza (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up sza and phi0\n", status);
    return status;
  }
  

  /**** Cloudless sky atmosphere ****/
  pmesg (" ... calling setup_atmosphere(), generating model atmosphere\n", input.verbose);
  status = setup_atmosphere (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up atmosphere\n", status);
    return status;
  }  


  /**** Setup temperature at computing levels ****/
  pmesg (" ... calling setup_temperature(), generating model temperature profile\n", input.verbose);
  status = setup_temperature (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up temperature\n", status);
    return status;
  }
  

  /**** Read cross section files, like O3, BrO, OclO, NO2, etc. ****/
  pmesg (" ... calling setup_crs(), reading cross section files\n", input.verbose);
  status = setup_crs (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up absorption cross sections\n", status);
    return status;
  }


  /**** Calculate Rayleigh scattering cross section ****/
  pmesg (" ... calling setup_rayleigh(), calculating Rayleigh scattering\n", input.verbose);
  status = setup_rayleigh (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up Rayleigh cross sections\n", status);
    return status;
  }


  /****  Optical properties of trace gases ****/
  pmesg (" ... calling setup_gases(), generating optical properties of trace gases\n", input.verbose);
  status = setup_gases (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up optical properties of trace gases\n", status);
    return status;
  }


  /****   Aerosol stuff   ****/
  pmesg (" ... calling setup_aerosols(), generating aerosols\n", input.verbose);
  status = setup_aerosol (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up aerosol properties\n", status);
    return status;
  }


  /**** Water cloud stuff ****/
  pmesg (" ... calling setup_wcloud(), generating water clouds\n", input.verbose);
  status = setup_wcloud (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up water cloud properties\n", status);
    return status;
  }

  /**** Ice cloud stuff ****/
  pmesg (" ... calling setup_icloud(), generating ice clouds\n", input.verbose);
  status = setup_icloud (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up ice cloud properties\n", status);
    return status;
  }


  /**** Redistribute profiles to required vertical resolution ****/
  pmesg (" ... calling setup_redistribute(), redistributing vertical profiles\n", input.verbose);
  status = setup_redistribute (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d doing vertical redistribution of optical properties\n", status);
    return status;
  }


  /****  Read extraterrestrial spectrum and correct for Earth-Sun distance ****/
  pmesg (" ... calling setup_extraterrestrial(), reading extraterrestrial spectrum\n", input.verbose);
  status = setup_extraterrestrial (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up extraterrestrial irradiance\n", status);
    return status;
  }


  /**** Setup wavelength dependent albedo ****/
  pmesg (" ... calling setup_albedo(), generating surface albedo\n", input.verbose);
  status = setup_albedo (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up wavelength dependent albedo\n", status);
    return status;
  }
  

  /**** Independent pixel stuff ****/
  pmesg (" ... calling setup_ipa(), generating independent pixels\n", input.verbose);
  status = setup_ipa (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up IPA properties\n", status);
    return status;
  }
  

  /****     Solve RTE     ****/
  pmesg (" ... calling solve_rte(), solving RTE\n", input.verbose);
  status = solve_rte (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d solving RTE\n", status);
    return status;
  }


  /**** Interpolate transmittance to high resolution wavelength grid ****/
  pmesg (" ... calling interpolate_transmittance(), interpolating transmittance\n", input.verbose);
  status = interpolate_transmittance (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating model output to hires wavelength grid\n", status);
    return status;
  }


  /**** Multiply with extraterrestrial irradiance ****/
  pmesg (" ... calling multiply_extraterrestrial()\n", input.verbose);
  status = multiply_extraterrestrial (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d multiplying with extraterrestrial irradiance\n", status);
    return status;
  }


  /**** Convolve with slit function ****/
  if (input.convolve) {
    pmesg (" ... calling convolve(), convoluting with slit function\n", input.verbose);
    status = convolve (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d convolving hires spectrum with slit function\n", status);
      return status;
    }
  }



  /**** Finally interpolate to output grid ****/
  if (input.spline) {
    pmesg (" ... calling spline_interpolate(), interpolating to output grid\n", input.verbose);
    status = spline_interpolate (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating spectrum to output grid\n", status);
      return status;
    }
  }

  
  /**** Wavelength-integrate 3D fields ****/
  /* if (input.rte.mc.passback3D && input.ck_scheme!=CK_NONE) { */
  if (input.rte.mc.passback3D) {
    status = integrate3D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d integrating 3D output over wavelength\n", status);
      return status;
    }
  }
  

  /**** Process 1D output ****/
  switch (input.processing) {
  case PROCESS_SUM:
    status = sum1D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d summing 1D output over wavelength\n", status);
      return status;
    }
    
    break;
    
  case PROCESS_NONE:
    break;
    
  case PROCESS_INT:
    status = integrate1D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d integrating 1D output over wavelength\n", status);
      return status;
    }
    
    break;

  case PROCESS_HEATING:
    status = heating1D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d calculating 1D heating/cooling rates\n", status);
      return status;
    }
    
    break;


  default:
    fprintf (stderr, "Error, unknown processing scheme %d\n", input.processing);
    return -1;
  }

  return 0;
}



/**************************************************************/
/* Check the model input for consistency.                     */
/**************************************************************/

int uvspec_check (input_struct Input)
{
  int i=0, status=0;
  int s1=0, s2=0;
  FILE *f=NULL;

  /* ***FIXME*** Still need to check (also in uvspec function):
     aerosol_haze
     aerosol_scale_ssa
     aerosol_set_gg
     aerosol_set_ssa
     aerosol_season
     aerosol_visibility
     aerosol_vulcan
     angstrom
     atmosphere_file
     data_files_path
     nstr
     rte_solver
     solar_file
     sza
     sza_file
     umu
     wc_file
     wc_set_gg
     wc_set_ssa
     wc_layer
     wvn
   */


  /* Albedo checks */
  
  if ( Input.atm.albedo > 1.0 ) {
    fprintf (stderr, "*** Surface albedo > 1.0: %f\n", Input.atm.albedo);
    status--;
  }
  else if ( Input.atm.albedo < 0.0 && strlen(Input.filename[FN_ALBEDO])==0 ) {
    fprintf (stderr, "*** Surface albedo < 0.0: %f\n", Input.atm.albedo);
    status--;
  }


  if (Input.brdf != BRDF_NONE && 
      (Input.rte.solver!=MONTECARLO && Input.rte.solver!=DISORT2)) {
    fprintf (stderr, "*** Error, BDRF is only implemented with MYSTIC and DISORT2!\n");
    status--;
  }


  if (Input.brdf == BRDF_CM && Input.rte.solver != DISORT2)  {
    fprintf (stderr, "*** Error, Cox and Munk currently only implemented for DISORT2\n");
    status--;
  }


  if ((strlen(Input.rte.mc.filename[FN_MC_RPV])>0) && 
      (Input.rte.solver!=MONTECARLO)) {
    fprintf (stderr, "*** Error, 2D RPV BDRF is only implemented with MYSTIC!\n");
    status--;
  }

  if ((strlen(Input.rte.mc.filename[FN_MC_RPV])>0  && strlen(Input.rte.mc.filename[FN_MC_RPVTYPE])==0) ||
      (strlen(Input.rte.mc.filename[FN_MC_RPV])==0 && strlen(Input.rte.mc.filename[FN_MC_RPVTYPE])>0)) {
    fprintf (stderr, "*** Error, you need to define both mc_rpv_file\n");
    fprintf (stderr, "***        and mc_rpv_type if you want 2D BRDF.\n");
    status--;
  }
  

  /* Day of year checks */

  if ( Input.day_of_year > 366 ) {
    fprintf (stderr, "*** Day of year > 366: %d\n", Input.day_of_year);
    status--;
  }
  else if ( Input.day_of_year < 0 && Input.day_of_year != -999 ) {
    fprintf (stderr, "*** Day of year < 0: %d\n", Input.day_of_year);
    status--;
  }

  /* Ozone column checks */

  if ( Input.atm.column[MOL_O3] < 0.0 && Input.atm.column[MOL_O3] != -999 ) {
    fprintf (stderr, "*** Ozone_column < 0.0: %f\n", Input.atm.column[MOL_O3]);
    status--;
  }


  /* Precipitable water checks */

  if ( Input.precip_water < 0.0 && Input.precip_water != -999 ) {
    fprintf (stderr, "*** Precipitable water < 0.0: %f\n", Input.precip_water);
    status--;
  }

  if ( Input.precip_water > 0.0 && Input.mixing_ratio[MX_H2O]>0) {
    fprintf (stderr, "*** It does not make sense to define both, h2o_mixing_ratio and h2o_precip\n");
    status--;
  }


  /* Aerosol checks */

  /* if any of the following four variables is set, all four must be set; */
  /* otherwise the information is incomplete!                             */
  if (Input.aer.seasn>0 || Input.aer.vulcan>0 || Input.aer.haze>0 || Input.aer.visibility>0) {

    if (Input.aer.seasn<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_season or aerosol_default!\n");
      status--;
    }

    if (Input.aer.vulcan<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_vulcan or aerosol_default!\n");
      status--;
    }

    if (Input.aer.haze<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_haze or aerosol_default!\n");
      status--;
    }

    if (Input.aer.visibility<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_visibility or aerosol_default!\n");
      status--;
    }
  }
  
  /* if no standard aerosol is defined, we don't want any of the aerosol varibles set */
  if (!Input.aer.standard && Input.aer.spec) {  
    fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_default or each\n");
    fprintf (stderr, "*** of aerosol_vulcan, aerosol_vulcan, aerosol_haze, and aerosol_visibility!\n"); 
    status--;
  }


  if ((Input.aer.seasn < 0 && Input.aer.seasn!=-999) || Input.aer.seasn >= AEROSOL_SEASON_NN) {
    fprintf (stderr, "*** Aerosol season out of range.\n");
    status--;
  }

  if ((Input.aer.haze < 0 && Input.aer.haze!=-999) || Input.aer.haze >= AEROSOL_HAZE_NN) {
    fprintf (stderr, "*** Aerosol haze out of range.\n");
    status--;
  }

  if ((Input.aer.vulcan < 0 && Input.aer.vulcan!=-999) || Input.aer.vulcan >= AEROSOL_VULCAN_NN) {
    fprintf (stderr, "*** Aerosol vulcan out of range.\n");
    status--;
  }
    

  s1=strlen(Input.aer.filename[FN_AER_SIZ]);
  s2=strlen(Input.aer.filename[FN_AER_REF]);

  if (s1>0 && (s2==0 && Input.aer.re==0)) {
    fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_refrac_index or aerosol_refrac_file\n");
    fprintf (stderr, "***                                 or omit the size distribution\n");
    status--;
  }

  if (s1==0 && (s2>0 || Input.aer.re>0)) {
    fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_sizedist_file\n");
    fprintf (stderr, "***                                 or omit the refractive index!\n");
    status--;
  }

  
  if (Input.wc.tau>=0.0 && Input.wc.tau550>=0.0) {
    fprintf (stderr, "*** It doesn't make sense to specify both wc_set_tau and wc_set_tau550;\n");
    fprintf (stderr, "*** please remove either one and retry!\n");
    status--;
  }

  if (Input.ic.tau>=0.0 && Input.ic.tau550>=0.0) {
    fprintf (stderr, "*** It doesn't make sense to specify both ic_set_tau and ic_set_tau550;\n");
    fprintf (stderr, "*** please remove either one and retry!\n");
    status--;
  }

  if (strlen(Input.filename[FN_WCLOUD])>0 && strlen(Input.filename[FN_WCLOUD_FILES])>0 ) {
    fprintf (stderr, "*** It doesn't make sense to specify both wc_file and wc_files;\n");
    fprintf (stderr, "*** please remove either one and retry!\n");
    status--;
  }

  /* need to specify a cloud field for the given cloud properties */
  if ((Input.wc.properties != WCPROP_NONE && Input.wc.properties != WCPROP_EXPLICIT) && 
      (strlen(Input.filename[FN_WCLOUD])==0 && strlen(Input.filename[FN_WCLOUD_IPA])==0)) {
    fprintf (stderr, "*** wc_properties only makes sense together with 'wc_file' or\n");
    fprintf (stderr, "*** wc_ipa_files!\n");
    status--;
  }
  
  /* need to specify a cloud field for the given cloud properties */
  if ((Input.ic.properties != ICPROP_NONE && Input.ic.properties != ICPROP_EXPLICIT) && 
      (strlen(Input.filename[FN_ICLOUD])==0 && strlen(Input.filename[FN_ICLOUD_IPA])==0)) {
    fprintf (stderr, "*** ic_properties only makes sense together with 'ic_file' or\n");
    fprintf (stderr, "*** ic_ipa_files!\n");
    status--;
  }


  if (strlen(Input.filename[FN_WCLOUD_PROP])>0 && !Input.wc.layer) {
    fprintf (stderr, "*** wc_properties requires optical properties defined per layer.\n");
    fprintf (stderr, "*** Please specify wc_layer and retry, if that is what you want!\n");
    status--;
  }


  if (strlen(Input.filename[FN_ICLOUD_PROP])>0 && !Input.ic.layer) {
    fprintf (stderr, "*** ic_properties requires optical properties defined per layer.\n");
    fprintf (stderr, "*** Please specify ic_layer and retry, if that is what you want!\n");
    status--;
  }

  /* RTE checks */

  switch (Input.rte.solver) {
  case SPSDISORT:
    if (Input.source==SRC_THERMAL) {
      fprintf (stderr, "*** spsdisort does not include thermal emission\n");
      status--;
    }

    break;

  case SDISORT:

    if (Input.source==SRC_THERMAL) {
      fprintf (stderr, "*** sdisort does not include thermal emission\n");
      status--;
    }

    if (Input.rte.nscat < 0 || Input.rte.nscat > 2) {
      fprintf (stderr, "*** nscat = %d is out of bounds\n", Input.rte.nscat);
      fprintf (stderr, "*** Only the values of 1 and 2 allowed when\n");
      fprintf (stderr, "*** combined with the sdisort solver.\n");
      status--;
    }

    if (Input.rte.nrefrac < 0 || Input.rte.nrefrac > 2) {
      fprintf (stderr, "*** nrefrac = %d is out of bounds\n", Input.rte.nrefrac);
      status--;
    }
    
    break;

  case TWOSTR:

    if (Input.rte.nprndis>0 )
      for (i=0;i<Input.rte.nprndis;i++)
	if ( !(Input.rte.prndis[i] == 1 || Input.rte.prndis[i] == 2) ) { 
	  fprintf (stderr, "*** twostr can only handle prndis set to 1 and/or 2\n");
	  status--;
	}

    break;

  case DISORT:
    break;

  case POLRADTRAN:
    break;

  case MONTECARLO:
    break;

  default:
    break;
  }


  if (Input.rte.mc.passback3D && Input.rte.mc.absorption!=MCABS_NONE && 
      !(Input.rte.mc.Nx_cloud * Input.rte.mc.Ny_cloud * Input.rte.mc.Nz_cloud)) {
    fprintf (stderr, "*** Error, need to specify mc_cloud_grid if mc_sample_grid and\n");
    fprintf (stderr, "***        mc_absorption are defined!\n");


    status--;
  }


  if (Input.rte.mc.absorption==MCABS_EMISSION && Input.source != SRC_THERMAL) {
    fprintf (stderr, "*** Error: mc_emission does only make sense with \"source thermal\"\n");
    status--;
  }


  /* check if zout is sorted in ascending order */
  for (i=1; i<Input.atm.nzout; i++)
    if (Input.atm.zout_sur[i] - Input.atm.zout_sur[i-1] < 0) {
      fprintf (stderr, "*** zout not sorted in ascending order\n");
      status--;
    }

  /* transmittance_wl_file and thermal_bands_file are mutually exclusive */
  if ((strlen(Input.filename[FN_WLTRANS])>0) &&
      (strlen(Input.filename[FN_WLBANDS])>0)) {
    fprintf (stderr, "*** Error, it does not make sense to specify both\n");
    fprintf (stderr, "*** transmittance_wl_file and thermal_bands_file!\n");
    status--;
  }

  /* solar_file and thermal_bands_file are mutually exclusive */
  if ((strlen(Input.filename[FN_EXTRATERRESTRIAL])>0) &&
      (strlen(Input.filename[FN_WLBANDS])>0)) {
    fprintf (stderr, "*** Error, it does not make sense to specify both\n");
    fprintf (stderr, "*** solar_file and thermal_bands_file!\n");
    status--;
  }

  /* Need to define a wavelength range if no correlated_k and no transmittance_wl_file */
  if ((Input.wl.start<0 || Input.wl.end<0) && 
      (Input.ck_scheme == CK_NONE || Input.ck_scheme == CK_SBDART) &&
      (strlen(Input.filename[FN_WLTRANS])==0) &&
      (strlen(Input.filename[FN_WLBANDS])==0)) {
    fprintf (stderr, "*** Error, need to define a wavelength range with 'wvn lower upper'!\n");
    status--;
  }

  /* wvn_index only makes sense with explicitely defined wavelength indices */
  if ((Input.wl.start_index>0 || Input.wl.end_index>0) && 
      ((Input.ck_scheme == CK_NONE || Input.ck_scheme == CK_SBDART) &&
       (strlen(Input.filename[FN_WLTRANS])==0) &&
       (strlen(Input.filename[FN_WLBANDS])==0))) {
    fprintf (stderr, "*** Error, wvn_index can only be used together with\n");
    fprintf (stderr, "*** explicitely defined wavelength grids!\n");
    status--;
  }

  /* wvn and wvn_index are mutually exclusive */
  if ((Input.wl.start_index>0 || Input.wl.end_index>0) && 
      (Input.wl.start>0 || Input.wl.end>0)) {
    fprintf (stderr, "*** Error, it does not make sense to define both 'wvn' and 'wvn_index'!\n");
    status--;
  }

  if ( Input.spline && !(strlen(Input.filename[FN_SPLINE])>0) ) {
    if ( Input.wl.start>0 && Input.wl.start > Input.spline_lambda_0 )  {
      fprintf (stderr, "*** Spline wavelength %f is smaller than wvn %f\n",
	       Input.spline_lambda_0, Input.wl.start);
      status--;
    }
    if ( Input.wl.end>0 && Input.wl.end < Input.spline_lambda_1 )  {
      fprintf (stderr, "*** Spline wavelength %f is greater than wvn %f\n",
	       Input.spline_lambda_1, Input.wl.end);
      status--;
    }
  }

  /* Check if slit_function file exists */
  if ( Input.convolve ) {
    if ( (f = fopen(Input.filename[FN_SLITFUNCTION], "r")) == NULL) {
      fprintf (stderr, "*** Can not read slit function file %s\n", Input.filename[FN_SLITFUNCTION]);
      status--;
    }
    else 
      fclose(f);
  }

  
  /* Cloud checks */
  if ( Input.cloudcover != -999.0 ) {
    if (Input.cloudcover < 0.0 || Input.cloudcover > 1.0) {
      fprintf (stderr, "*** wc_cloudcover not between 0 and 1: %f\n", Input.cloudcover);
      status--;
    }
    if (strlen(Input.filename[FN_WCLOUD])==0 ) {
      fprintf (stderr, "*** wc_cloudcover must be accompanied by a cloud file specified by wc_file\n");
      status--;
    }
  }

  
  /* ck and libnetcdf */ 
#if HAVE_LIBNETCDF
#else
  if (Input.ck_scheme == CK_KATO) {
    fprintf (stderr, "   **********************************************************************\n");
    fprintf (stderr, "   * You have built uvspec without libnetcdf and hence may not          *\n");
    fprintf (stderr, "   * use the 'correlated_k kato' option. Please get netcdf and rebuild. *\n");
    fprintf (stderr, "   **********************************************************************\n");
    status--;
  }
#endif


  /* The MYSTIC solver */ 
#if HAVE_MYSTIC
#else
  if (Input.rte.solver == MONTECARLO) {
    fprintf (stderr, "   ********************************************************************\n");
    fprintf (stderr, "   * You have built uvspec without the three-dimensional MYSTIC       *\n");
    fprintf (stderr, "   * module. MYSTIC is currently not released to the public. If you   *\n");
    fprintf (stderr, "   * have an interest in three-dimensional calculations and if you    *\n");
    fprintf (stderr, "   * have a reasonable project in mind, please contact                *\n");
    fprintf (stderr, "   * bernhard.mayer@dlr.de.                                           *\n");
    fprintf (stderr, "   ********************************************************************\n");
    status--;
  }
#endif


  /* Fu and Liou */ 
#if HAVE_FULIOU
#else
  if (Input.ck_scheme == CK_FU) {
    fprintf (stderr, "   **********************************************************************\n");
    fprintf (stderr, "   * You have built uvspec without support for the Fu and Liou          *\n");
    fprintf (stderr, "   * parameterization. If you are interested, please contact the        *\n");
    fprintf (stderr, "   * libRadtran authors.                                                *\n");
    fprintf (stderr, "   **********************************************************************\n");
    status--;
  }
#endif


  /* Fu and Liou at present only works with water clouds defined per layer */
  if (Input.ck_scheme == CK_FU && strlen(Input.filename[FN_WCLOUD])>0 && !Input.wc.layer) {
    fprintf (stderr, "   ***********************************************************************\n");
    fprintf (stderr, "   *** The Fu and Liou c-k parameterization only works with water cloud  *\n");
    fprintf (stderr, "   *** properties defined per layer. Please add 'wc_layer' to the input  *\n");
    fprintf (stderr, "   *** file and check the documentation for its meaning!                 *\n");
    fprintf (stderr, "   ***********************************************************************\n");
    status--;
  }

  /* Fu and Liou at present only works with ice clouds defined per layer */
  if (Input.ck_scheme == CK_FU && strlen(Input.filename[FN_ICLOUD])>0 && !Input.ic.layer) {
    fprintf (stderr, "   ***********************************************************************\n");
    fprintf (stderr, "   *** The Fu and Liou c-k parameterization only works with ice cloud    *\n");
    fprintf (stderr, "   *** properties defined per layer. Please add 'ic_layer' to the input  *\n");
    fprintf (stderr, "   *** file and check the documentation for its meaning!                 *\n");
    fprintf (stderr, "   ***********************************************************************\n");
    status--;
  }


  /* Yang et al. at present only works with ice clouds defined per layer */
  if ((Input.ic.properties==ICPROP_YANG || 
       Input.ic.properties==ICPROP_FU) && !Input.ic.layer) {
    fprintf (stderr, "   ***********************************************************************\n");
    fprintf (stderr, "   *** The Yang et al. and Fu et al. ice cloud parameterizations only    *\n");
    fprintf (stderr, "   *** work with cloud properties defined per layer. Please add          *\n");
    fprintf (stderr, "   *** 'ic_layer' to the input file and check the documentation for      *\n");
    fprintf (stderr, "   *** its meaning!                                                      *\n");
    fprintf (stderr, "   ***********************************************************************\n");
    status--;
  }
  

  if ((Input.ic.properties==ICPROP_YANG) && Input.ic.habit<0) {
    fprintf (stderr, "Error, unknown ice crystal habit. Check ic_habit!\n");
    status--;
  }


  /* SBDART */ 
#if HAVE_SBDART
  if (Input.ck_scheme == CK_SBDART && !Input.quiet) {
    fprintf (stderr, "   **********************************************************************\n");
    fprintf (stderr, "   * When using the SBDART gas absorption for any result that goes into *\n");
    fprintf (stderr, "   * a publication, please don't forget to reference SBDART properly!   *\n");
    fprintf (stderr, "   **********************************************************************\n");
  }
#else
  if (Input.ck_scheme == CK_SBDART) {
    fprintf (stderr, "   **********************************************************************\n");
    fprintf (stderr, "   * You have built uvspec without support for the SBDART gas           *\n");
    fprintf (stderr, "   * parameterization. If you are interested, please contact the        *\n");
    fprintf (stderr, "   * libRadtran authors.                                                *\n");
    fprintf (stderr, "   **********************************************************************\n");
    status--;
  }
#endif

  /* Output checks */
  if ( Input.processing ==  PROCESS_HEATING ) {
    if ( Input.rte.solver != TWOSTR &&
	 Input.rte.solver != DISORT &&
	 Input.rte.solver != SDISORT &&
	 Input.rte.solver != DISORT2 &&
	 Input.rte.solver != SPSDISORT &&
	 Input.rte.solver != POLRADTRAN ) {
      fprintf (stderr, "   The heating output option may only be used with the disort, \n");
      fprintf (stderr, "   disort2, sdisort, twostr, spsdisort and polradtran solvers. \n");
      status--;
    }
  }

  
  return status;
}





void pmesg (char *message, int verbose)
{
  if (verbose) {
    fprintf (stderr, message);
    fflush (stderr);
  }
}
